from __future__ import annotations

import asyncio
import contextlib
import datetime
import itertools
import operator
import os.path
import pathlib
import random
from collections.abc import AsyncIterator
from types import MethodType
from typing import Callable

import aiohttp
import aiohttp_client_cache
import asyncstdlib
import discord
import ujson
from apscheduler.schedulers.asyncio import AsyncIOScheduler
from asyncspotify import Client as SpotifyClient
from asyncspotify import ClientCredentialsFlow
from discord.abc import Messageable
from discord.ext.commands import Context
from discord.types.embed import EmbedType
from pytz import utc

from pylav._config import __VERSION__, CONFIG_DIR
from pylav._logging import getLogger
from pylav.dispatcher import DispatchManager
from pylav.envvars import (
    EXTERNAL_UNMANAGED_HOST,
    EXTERNAL_UNMANAGED_PASSWORD,
    EXTERNAL_UNMANAGED_PORT,
    EXTERNAL_UNMANAGED_SSL,
    JAVA_EXECUTABLE,
    REDIS_FULL_ADDRESS_RESPONSE_CACHE,
    TASK_TIMER_UPDATE_BUNDLED_EXTERNAL_PLAYLISTS_DAYS,
    TASK_TIMER_UPDATE_BUNDLED_PLAYLISTS_DAYS,
    TASK_TIMER_UPDATE_EXTERNAL_PLAYLISTS_DAYS,
    USE_BUNDLED_EXTERNAL_LAVA_LINK_NODE,
    USE_BUNDLED_EXTERNAL_PYLAV_NODE,
)
from pylav.events import Event
from pylav.exceptions import (
    AnotherClientAlreadyRegistered,
    NoNodeAvailable,
    NoNodeWithRequestFunctionalityAvailable,
    PyLavInvalidArguments,
)
from pylav.m3u8_parser import M3U8Parser
from pylav.managed_node import LocalNodeManager
from pylav.node import Node
from pylav.node_manager import NodeManager
from pylav.player import Player
from pylav.player_manager import PlayerManager
from pylav.query import MAX_RECURSION_DEPTH, Query
from pylav.radio import RadioBrowser
from pylav.sql.aiohttp_postgres_cache import PostgresCacheBackend
from pylav.sql.clients.equalizer_manager import EqualizerConfigManager
from pylav.sql.clients.lib import LibConfigManager
from pylav.sql.clients.nodes_db_manager import NodeConfigManager
from pylav.sql.clients.player_config_manager import PlayerConfigManager
from pylav.sql.clients.player_state_db_manager import PlayerStateDBManager
from pylav.sql.clients.playlist_manager import PlaylistConfigManager
from pylav.sql.clients.query_manager import QueryCacheManager
from pylav.sql.clients.updater import UpdateSchemaManager
from pylav.sql.models import LibConfigModel
from pylav.tracks import Track
from pylav.types import BotT, CogT, ContextT, InteractionT, LavalinkResponseT
from pylav.utils import PyLavContext, SingletonMethods, _get_context, _process_commands, _Singleton, add_property
from pylav.vendored import aiopath

try:
    from redbot.core.i18n import Translator

    _ = Translator("PyLavPlayer", pathlib.Path(__file__))
except ImportError:
    _ = lambda x: x

LOGGER = getLogger("PyLav.Client")


class Client(metaclass=_Singleton):
    """
    Represents a Lavalink client used to manage nodes and connections.

    .. _event loop: https://docs.python.org/3/library/asyncio-eventloop.html

    Parameters
    ----------
    bot : :class:`BotT`
        The bot instance.
    cog: :class:`CogT`
        The cog to register the client to.
    player: Optional[:class:`Player`]
        The class that should be used for the player. Defaults to ``Player``.
        Do not change this unless you know what you are doing!
    connect_back: Optional[:class:`bool`]
        A boolean that determines if a player will connect back to the
        node it was originally connected to. This is not recommended doing since
        the player will most likely be performing better in the new node. Defaults to `False`.

        Warning
        -------
        If this option is enabled and the player's node is changed through `Player.change_node` after
        the player was moved via the fail-over mechanism, the player will still move back to the original
        node when it becomes available. This behaviour can be avoided in custom player implementations by
        setting `self._original_node` to `None` in the `change_node` function.
    config_folder: Optional[:class:`pathlib.Path`]
        The path to the config folder. Defaults to ``CONFIG_DIR``.
    """

    _local_node_manager: LocalNodeManager

    _asyncio_lock = asyncio.Lock()
    _config: LibConfigModel
    __cogs_registered = set()
    __old_process_command_method: Callable = None  # type: ignore
    __old_get_context: Callable = None  # type: ignore
    _initiated = False

    def __init__(
        self,
        bot: BotT,
        cog: CogT,
        player: type[Player] = Player,  # type: ignore
        connect_back: bool = False,
        config_folder: aiopath.AsyncPath | pathlib.Path = CONFIG_DIR,
    ):
        try:
            setattr(bot, "_pylav_client", self)
            add_property(bot, "lavalink", lambda b: b._pylav_client)
            add_property(bot, "pylav", lambda b: b._pylav_client)
            if self.__old_process_command_method is None:
                self.__old_process_command_method = bot.process_commands
            if self.__old_get_context is None:
                self.__old_get_context = bot.get_context
            self.ready = asyncio.Event()
            bot.process_commands = MethodType(_process_commands, bot)
            bot.get_context = MethodType(_get_context, bot)
            self.__cogs_registered.add(cog.__cog_name__)
            config_folder = pathlib.Path(config_folder)
            (config_folder / ".data").mkdir(exist_ok=True, parents=True)
            self._config_folder = aiopath.AsyncPath(config_folder)
            self._bot = bot
            self._user_id = str(bot.user.id)
            if REDIS_FULL_ADDRESS_RESPONSE_CACHE:
                self._aiohttp_client_cache = aiohttp_client_cache.RedisBackend(
                    address=REDIS_FULL_ADDRESS_RESPONSE_CACHE,
                    cache_control=True,
                    allowed_codes=(200,),
                    allowed_methods=("GET",),
                    include_headers=True,
                    expire_after=datetime.timedelta(days=1),
                    timeout=2.5,
                    ignored_params=["auth_token", "timestamp"],
                )
            else:
                self._aiohttp_client_cache = PostgresCacheBackend(
                    cache_control=True,
                    allowed_codes=(200,),
                    allowed_methods=("GET",),
                    include_headers=True,
                    ignored_params=["auth_token", "timestamp"],
                    expire_after=datetime.timedelta(days=1),
                    timeout=2.5,
                )
            self._session = aiohttp.ClientSession(timeout=aiohttp.ClientTimeout(total=30), json_serialize=ujson.dumps)
            self._cached_session = aiohttp_client_cache.CachedSession(
                timeout=aiohttp.ClientTimeout(total=30), json_serialize=ujson.dumps, cache=self._aiohttp_client_cache
            )
            self._node_manager = NodeManager(
                self,
                external_host=EXTERNAL_UNMANAGED_HOST,
                external_port=EXTERNAL_UNMANAGED_PORT,
                external_password=EXTERNAL_UNMANAGED_PASSWORD,
                external_ssl=EXTERNAL_UNMANAGED_SSL,
            )
            self._player_manager = PlayerManager(self, player)
            self._lib_config_manager = LibConfigManager(self)
            self._node_config_manager = NodeConfigManager(self)
            self._playlist_config_manager = PlaylistConfigManager(self)
            self._query_cache_manager = QueryCacheManager(self)
            self._update_schema_manager = UpdateSchemaManager(self)
            self._dispatch_manager = DispatchManager(self)
            self._player_state_db_manager = PlayerStateDBManager(self)
            self._player_config_manager = PlayerConfigManager(self)
            self._equalizer_config_manager = EqualizerConfigManager(self)
            self._radio_manager = RadioBrowser(self)
            self._m3u8parser = M3U8Parser(self)
            self._connect_back = connect_back
            self._warned_about_no_search_nodes = False
            self._spotify_client_id = None
            self._spotify_client_secret = None
            self._spotify_auth = None
            self._shutting_down = False
            self.enable_managed_node = None
            self._scheduler = AsyncIOScheduler()
            self._scheduler.configure(timezone=utc)
        except Exception:
            LOGGER.exception("Failed to initialize Lavalink")
            raise

    async def wait_until_ready(self, timeout: float | None = None) -> None:
        await asyncio.wait_for(self.ready.wait(), timeout=timeout)

    @property
    def initialized(self) -> bool:
        """Returns whether the client has been initialized."""
        return self._initiated

    @property
    def scheduler(self) -> AsyncIOScheduler:
        """Returns the scheduler."""
        return self._scheduler

    @property
    def radio_browser(self) -> RadioBrowser:
        """Returns the radio browser instance."""
        return self._radio_manager

    @property
    def player_config_manager(self) -> PlayerConfigManager:
        """Returns the player config manager."""
        return self._player_config_manager

    @property
    def spotify_client(self) -> SpotifyClient:
        """Returns the spotify client."""
        return SpotifyClient(self._spotify_auth)

    @property
    def is_shutting_down(self) -> bool:
        """Returns whether the client is shutting down."""
        return self._shutting_down

    @property
    def node_db_manager(self) -> NodeConfigManager:
        """Returns the sql node config manager."""
        return self._node_config_manager

    @property
    def player_state_db_manager(self) -> PlayerStateDBManager:
        return self._player_state_db_manager

    @property
    def playlist_db_manager(self) -> PlaylistConfigManager:
        """Returns the sql playlist config manager."""
        return self._playlist_config_manager

    @property
    def equalizer_db_manager(self) -> EqualizerConfigManager:
        """Returns the sql equalizer config manager."""
        return self._equalizer_config_manager

    @property
    def lib_db_manager(self) -> LibConfigManager:
        """Returns the sql lib config manager."""
        return self._lib_config_manager

    @property
    def query_cache_manager(self) -> QueryCacheManager:
        """Returns the query cache manager."""
        return self._query_cache_manager

    @property
    def managed_node_controller(self) -> LocalNodeManager:
        return self._local_node_manager

    @property
    def node_manager(self) -> NodeManager:
        return self._node_manager

    @property
    def player_manager(self) -> PlayerManager:
        return self._player_manager

    @property
    def config_folder(self) -> aiopath.AsyncPath:
        return self._config_folder

    @property
    def bot(self) -> BotT:
        return self._bot

    @property
    def session(self) -> aiohttp.ClientSession:
        return self._session

    @property
    def cached_session(self) -> aiohttp_client_cache.CachedSession:
        return self._cached_session

    @property
    def lib_version(self) -> str:
        return __VERSION__

    @property
    def bot_id(self) -> str:
        return self._user_id

    @SingletonMethods.run_once_async
    async def initialize(
        self,
    ) -> None:  # sourcery no-metrics
        try:
            if not self._initiated:
                async with self._asyncio_lock:
                    if not self._initiated:
                        self._initiated = True
                        self.ready.clear()
                        await self.bot.wait_until_ready()
                        if hasattr(self.bot, "get_shared_api_tokens") and callable(
                            getattr(self.bot, "get_shared_api_tokens")
                        ):
                            spotify = await self.bot.get_shared_api_tokens("spotify")
                            client_id = spotify.get("client_id")
                            client_secret = spotify.get("client_secret")
                            LOGGER.debug(
                                "Existing Spotify tokens found; Using them - %s - %s", client_id, client_secret
                            )
                        else:
                            LOGGER.info("PyLav being run from a non Red bot")
                            client_id = None
                            client_secret = None
                        await self._lib_config_manager.initialize()
                        await self._update_schema_manager.run_updates()
                        await self._radio_manager.initialize()
                        await self._player_manager.initialize()

                        self._config = await self._lib_config_manager.get_config(
                            config_folder=self._config_folder,
                            java_path=JAVA_EXECUTABLE,
                            enable_managed_node=True,
                            auto_update_managed_nodes=True,
                            localtrack_folder=self._config_folder / "music",
                            use_bundled_lava_link_external=USE_BUNDLED_EXTERNAL_LAVA_LINK_NODE,
                            use_bundled_pylav_external=USE_BUNDLED_EXTERNAL_PYLAV_NODE,
                        )
                        if self._config.java_path != JAVA_EXECUTABLE:
                            if JAVA_EXECUTABLE != "java" and os.path.exists(JAVA_EXECUTABLE):
                                self._config.java_path = JAVA_EXECUTABLE
                                await self._config.save()
                        LOGGER.info("Config folder: %s", self._config.config_folder)
                        LOGGER.info("Localtracks folder: %s", self._config.localtrack_folder)
                        auto_update_managed_nodes = self._config.auto_update_managed_nodes
                        self.enable_managed_node = self._config.enable_managed_node
                        self._config_folder = aiopath.AsyncPath(self._config.config_folder)
                        localtrack_folder = aiopath.AsyncPath(self._config.localtrack_folder)
                        data = await self._node_config_manager.get_bundled_node_config()
                        if not await asyncstdlib.all([client_id, client_secret]):
                            spotify_data = data.yaml["plugins"]["topissourcemanagers"]["spotify"]
                            client_id = spotify_data["clientId"]
                            client_secret = spotify_data["clientSecret"]
                        elif await asyncstdlib.all([client_id, client_secret]):
                            if (
                                data.yaml["plugins"]["topissourcemanagers"]["spotify"]["clientId"] != client_id
                                or data.yaml["plugins"]["topissourcemanagers"]["spotify"]["clientSecret"]
                                != client_secret
                            ):
                                data.yaml["plugins"]["topissourcemanagers"]["spotify"]["clientId"] = client_id
                                data.yaml["plugins"]["topissourcemanagers"]["spotify"]["clientSecret"] = client_secret
                            await data.save()
                        self._spotify_client_id = client_id
                        self._spotify_client_secret = client_secret
                        self._spotify_auth = ClientCredentialsFlow(
                            client_id=self._spotify_client_id, client_secret=self._spotify_client_secret
                        )
                        from pylav.localfiles import LocalFile

                        await LocalFile.add_root_folder(path=localtrack_folder, create=True)
                        self._user_id = str(self._bot.user.id)
                        self._local_node_manager = LocalNodeManager(self, auto_update=auto_update_managed_nodes)

                        if (
                            not (
                                self._node_manager._unmanaged_external_password
                                and self._node_manager._unmanaged_external_host
                            )
                        ) and self.enable_managed_node:
                            await self._local_node_manager.start(java_path=self._config.java_path)
                        else:
                            self._local_node_manager.ready.set()
                        await self.node_manager.connect_to_all_nodes()
                        await self.node_manager.wait_until_ready()
                        await self.player_manager.restore_player_states()
                        time_now = datetime.datetime.now(tz=datetime.timezone.utc)
                        if self._config.next_execution_update_bundled_playlists is None:
                            self._config.next_execution_update_bundled_playlists = time_now + datetime.timedelta(
                                minutes=5
                            )
                        if self._config.next_execution_update_bundled_external_playlists is None:
                            self._config.next_execution_update_bundled_external_playlists = (
                                time_now + datetime.timedelta(minutes=10)
                            )
                        if self._config.next_execution_update_external_playlists is None:
                            self._config.next_execution_update_external_playlists = time_now + datetime.timedelta(
                                minutes=30
                            )
                        self._scheduler.add_job(
                            self._query_cache_manager.delete_old,
                            trigger="interval",
                            seconds=600,
                            max_instances=1,
                            replace_existing=True,
                            name="cache_delete_old",
                            coalesce=True,
                            id=f"{self.bot.user.id}-cache_delete_old",
                        )
                        self._scheduler.add_job(
                            self.playlist_db_manager.update_bundled_playlists,
                            trigger="interval",
                            days=TASK_TIMER_UPDATE_BUNDLED_PLAYLISTS_DAYS,
                            max_instances=1,
                            next_run_time=self._config.next_execution_update_bundled_playlists,
                            replace_existing=True,
                            name="update_bundled_playlists",
                            coalesce=True,
                            id=f"{self.bot.user.id}-update_bundled_playlists",
                            misfire_grace_time=None,
                        )
                        LOGGER.info(
                            "Scheduling first run of Bundled Playlist update task to: %s",
                            self._config.next_execution_update_bundled_playlists,
                        )
                        self._scheduler.add_job(
                            self.playlist_db_manager.update_bundled_external_playlists,
                            trigger="interval",
                            days=TASK_TIMER_UPDATE_BUNDLED_EXTERNAL_PLAYLISTS_DAYS,
                            max_instances=1,
                            next_run_time=self._config.next_execution_update_bundled_external_playlists,
                            replace_existing=True,
                            name="update_bundled_external_playlists",
                            coalesce=True,
                            id=f"{self.bot.user.id}-update_bundled_external_playlists",
                            misfire_grace_time=None,
                        )
                        LOGGER.info(
                            "Scheduling first run of Bundled External Playlist update task to: %s",
                            self._config.next_execution_update_bundled_external_playlists,
                        )
                        self._scheduler.add_job(
                            self.playlist_db_manager.update_external_playlists,
                            trigger="interval",
                            days=TASK_TIMER_UPDATE_EXTERNAL_PLAYLISTS_DAYS,
                            max_instances=1,
                            next_run_time=self._config.next_execution_update_external_playlists,
                            replace_existing=True,
                            name="update_external_playlists",
                            coalesce=True,
                            id=f"{self.bot.user.id}-update_external_playlists",
                            misfire_grace_time=None,
                        )
                        LOGGER.info(
                            "Scheduling first run of External Playlist update task to: %s",
                            self._config.next_execution_update_external_playlists,
                        )
                        await self._config.save()
                        self._scheduler.start()
                        self.ready.set()
                        LOGGER.info("PyLav is ready.")
        except Exception as exc:
            LOGGER.critical("Failed start up", exc_info=exc)
            raise exc

    async def register(self, cog: CogT) -> None:
        LOGGER.info("Registering cog %s", cog.__cog_name__)
        if (instance := getattr(self.bot, "lavalink", None)) and not isinstance(instance, Client):
            raise AnotherClientAlreadyRegistered(
                f"Another client instance has already been registered to bot.lavalink with type: {type(instance)}"
            )
        self.__cogs_registered.add(cog.__cog_name__)

    async def update_spotify_tokens(self, client_id: str, client_secret: str) -> None:
        LOGGER.info("Updating Spotify Tokens")
        LOGGER.debug("New Spotify token: ClientId: %s || ClientSecret: %s", client_id, client_secret)
        self._spotify_client_id = client_id
        self._spotify_client_secret = client_secret
        self._spotify_auth = ClientCredentialsFlow(
            client_id=self._spotify_client_id, client_secret=self._spotify_client_secret
        )
        bundled_node_config = await self._node_config_manager.get_bundled_node_config()
        bundled_node_config.yaml["plugins"]["topissourcemanagers"]["spotify"]["clientId"] = client_id
        bundled_node_config.yaml["plugins"]["topissourcemanagers"]["spotify"]["clientSecret"] = client_secret
        await bundled_node_config.save()

    async def add_node(
        self,
        *,
        unique_identifier: int,
        host: str,
        port: int,
        password: str,
        resume_key: str = None,
        resume_timeout: int = 60,
        name: str = None,
        reconnect_attempts: int = -1,
        ssl: bool = False,
        search_only: bool = False,
        managed: bool = False,
        skip_db: bool = False,
        yaml: dict | None = None,
        disabled_sources: list[str] = None,
        extras: dict = None,
    ) -> Node:
        """
        Adds a node to Lavalink's node manager.

        Parameters
        ----------
        host: :class:`str`
            The address of the Lavalink node.
        port: :class:`int`
            The port to use for websocket and REST connections.
        password: :class:`str`
            The password used for authentication.
        resume_key: Optional[:class:`str`]
            A resume key used for resuming a session upon re-establishing a WebSocket connection to Lavalink.
            Defaults to `None`.
        resume_timeout: Optional[:class:`int`]
            How long the node should wait for a connection while disconnected before clearing all players.
            Defaults to `60`.
        name: :class:`str`
            An identifier for the node that will show in logs. Defaults to `None`
        reconnect_attempts: Optional[:class:`int`]
            The amount of times connection with the node will be reattempted before giving up.
            Set to `-1` for infinite. Defaults to `3`.
        ssl: Optional[:class:`bool`]
            Whether to use SSL for the connection. Defaults to `False`.
        search_only: :class:`bool`
            Whether the node should only be used for searching. Defaults to `False`.
        unique_identifier: :class:`in`
            A unique identifier for the node. Defaults to `None`.
        skip_db: :class:`bool`
            Whether the node should skip the database. Defaults to `False`.
        yaml: Optional[:class:`dict`]
            A dictionary of extra information to be stored in the node. Defaults to `None`.
        extras: Optional[:class:`dict`]
            A dictionary of extra information to be stored in the node. Defaults to `None`.
        managed: :class:`bool`
            Whether the node is managed by the client. Defaults to `False`.
        disabled_sources: Optional[:class:`list`[:class:`str`]]
            A list of sources that should be disabled for the node. Defaults to `None`.
        """
        return await self.node_manager.add_node(
            host=host,
            port=port,
            password=password,
            resume_key=resume_key,
            resume_timeout=resume_timeout,
            name=name,
            reconnect_attempts=reconnect_attempts,
            ssl=ssl,
            search_only=search_only,
            unique_identifier=unique_identifier,
            skip_db=skip_db,
            managed=managed,
            yaml=yaml,
            disabled_sources=disabled_sources,
            extras=extras or {},
        )

    async def decode_track(self, track: str, feature: str = None) -> dict | None:
        """|coro|
        Decodes a base64-encoded track string into a dict.

        Parameters
        ----------
        track: :class:`str`
            The base64-encoded `track` string.
        feature: Optional[:class:`str`]
            The feature to decode the track for. Defaults to `None`.

        Returns
        -------
        :class:`dict`
            A dict representing the track's information.
        """
        if not self.node_manager.available_nodes:
            raise NoNodeAvailable(_("No available nodes!"))
        node = await self.node_manager.find_best_node(feature=feature)
        if node is None and feature:
            raise NoNodeWithRequestFunctionalityAvailable(
                _("No node with {feature} functionality available!").format(feature=feature), feature=feature
            )
        return await node.decode_track(track)

    async def decode_tracks(self, tracks: list, feature: str = None) -> list[dict]:
        """|coro|
        Decodes a list of base64-encoded track strings into a dict.

        Parameters
        ----------
        tracks: list[:class:`str`]
            A list of base64-encoded `track` strings.
        feature: Optional[:class:`str`]
            The feature to decode the tracks for. Defaults to `None`.

        Returns
        -------
        List[:class:`dict`]
            A list of dicts representing track information.
        """
        if not self.node_manager.available_nodes:
            raise NoNodeAvailable(_("No available nodes!"))
        node = await self.node_manager.find_best_node(feature=feature)
        if node is None and feature:
            raise NoNodeWithRequestFunctionalityAvailable(
                _("No node with {feature} functionality available!").format(feature=feature), feature=feature
            )
        return await node.decode_tracks(tracks)

    @staticmethod
    async def routeplanner_status(node: Node) -> dict | None:
        """|coro|
        Gets the route-planner status of the target node.

        Parameters
        ----------
        node: :class:`Node`
            The node to use for the query.

        Returns
        -------
        :class:`dict`
            A dict representing the route-planner information.
        """
        return await node.routeplanner_status()

    @staticmethod
    async def routeplanner_free_address(node: Node, address: str) -> bool:
        """|coro|
        Gets the route-planner status of the target node.

        Parameters
        ----------
        node: :class:`Node`
            The node to use for the query.
        address: :class:`str`
            The address to free.

        Returns
        -------
        :class:`bool`
            True if the address was freed, False otherwise.
        """
        return await node.routeplanner_free_address(address)

    @staticmethod
    async def routeplanner_free_all_failing(node: Node) -> bool:
        """|coro|
        Gets the route-planner status of the target node.

        Parameters
        ----------
        node: :class:`Node`
            The node to use for the query.

        Returns
        -------
        :class:`bool`
            True if all failing addresses were freed, False otherwise.
        """
        return await node.routeplanner_free_all_failing()

    def dispatch_event(self, event: Event):
        asyncio.create_task(self._dispatch_event(event))

    async def _dispatch_event(self, event: Event):
        """|coro|
        Dispatches the given event to all registered hooks.

        Parameters
        ----------
        event: :class:`Event`
            The event to dispatch to the hooks.
        """
        event_dispatcher = [self._dispatch_manager.dispatch]

        task_list = []
        for hook in itertools.chain(
            event_dispatcher,
        ):
            task = asyncio.create_task(hook(event))  # type: ignore
            task.set_name(f"Event hook {hook.__name__}")
            task.add_done_callback(self.__done_callback)
            task_list.append(task)
        await asyncio.gather(*task_list)

    @staticmethod
    def __done_callback(task: asyncio.Task):
        exc = task.exception()
        if exc is not None:
            name = task.get_name()
            LOGGER.warning("Event hook %s encountered an exception!", name)
            LOGGER.debug("Event hook %s encountered an exception!", name, exc_info=exc)

    async def unregister(self, cog: discord.ext.commands.Cog):
        """|coro|
        Unregister the specified Cog and if no cogs are left closes the client.

        Parameters
        ----------
        cog: :class:`discord.ext.commands.Cog`
            The cog to unregister.
        """
        if self._shutting_down:
            return
        async with self._asyncio_lock:
            if not self._shutting_down:
                self.__cogs_registered.discard(cog.__cog_name__)
                LOGGER.info("%s has been unregistered", cog.__cog_name__)
                if not self.__cogs_registered:
                    self._shutting_down = True
                    self.ready.clear()
                    try:
                        Client._instances.clear()
                        SingletonMethods.reset()
                        self._initiated = False
                        await self.player_manager.save_all_players()
                        await self.player_manager.shutdown()
                        await self._node_manager.close()
                        await self._local_node_manager.shutdown()
                        await self._session.close()
                        await self._cached_session.close()
                        self._scheduler.shutdown(wait=True)
                    except Exception as e:
                        LOGGER.critical("Failed to shutdown the client", exc_info=e)
                    if self.__old_process_command_method is not None:
                        self.bot.process_commands = self.__old_process_command_method
                    if self.__old_get_context is not None:
                        self.bot.get_context = self.__old_get_context
                    del self.bot._pylav_client  # noqa
                    LOGGER.info("All cogs have been unregistered, PyLav client has been shutdown.")

    def get_player(self, guild: discord.Guild | int | None) -> Player | None:
        """Gets the player for the target guild.

        Parameters
        ----------
        guild: :class:`discord.Guild`
            The guild to get the player for.

        Returns
        -------
        :class:`Player`
            The player for the target guild.
        """
        if not guild:
            return None
        if not isinstance(guild, int):
            guild = guild.id
        return self.player_manager.get(guild)

    async def connect_player(
        self,
        requester: discord.Member,
        channel: discord.channel.VocalGuildChannel,
        node: Node = None,
        self_deaf: bool = True,
    ) -> Player:
        """|coro|
        Connects the player for the target guild.

        Parameters
        ----------
        channel: :class:`discord.channel.VocalGuildChannel`
            The channel to connect to.
        node: :class:`Node`
            The node to use for the connection.
        self_deaf: :class:`bool`
            Whether the bot should be deafened.
        requester: :class:`discord.Member`
            The member requesting the connection.
        Returns
        -------
        :class:`Player`
            The player for the target guild.
        """
        p = await self.player_manager.create(channel, channel.rtc_region, node, self_deaf, requester)
        return p

    async def construct_embed(
        self,
        *,
        embed: discord.Embed = None,
        colour: discord.Colour | int | None = None,
        color: discord.Colour | int | None = None,
        title: str = None,
        type: EmbedType = "rich",
        url: str = None,
        description: str = None,
        timestamp: datetime.datetime = None,
        author_name: str = None,
        author_url: str = None,
        thumbnail: str = None,
        footer: str = None,
        footer_url: str = None,
        messageable: Messageable | InteractionT = None,
    ) -> discord.Embed:

        if messageable and not colour and not color and hasattr(self._bot, "get_embed_color"):
            colour = await self._bot.get_embed_color(messageable)
        elif colour or color:
            colour = colour or color
        if timestamp and isinstance(timestamp, datetime.datetime):
            timestamp = timestamp
        else:
            timestamp = datetime.datetime.now(tz=datetime.timezone.utc)
        contents = dict(
            title=title,
            type=type,
            url=url,
            description=description,
            timestamp=timestamp.isoformat(),
        )
        embed = embed.to_dict() if embed is not None else {}
        contents |= embed
        new_embed = discord.Embed.from_dict(contents)
        new_embed.color = colour

        if footer:
            new_embed.set_footer(text=footer, icon_url=footer_url)
        if thumbnail:
            new_embed.set_thumbnail(url=thumbnail)
        if author_url and author_name:
            new_embed.set_author(name=author_name, icon_url=author_url)
        return new_embed

    async def get_context(self, what: discord.Message | ContextT | InteractionT) -> PyLavContext:
        if isinstance(what, PyLavContext):
            return what
        elif isinstance(what, Context):
            ctx_ = what.interaction or what.message
            ctx: PyLavContext = await self._bot.get_context(ctx_, cls=PyLavContext)  # type: ignore

        else:
            ctx: PyLavContext = await self._bot.get_context(what, cls=PyLavContext)  # type: ignore
        return ctx

    async def update_localtracks_folder(self, folder: str | None) -> None:
        from pylav.localfiles import LocalFile

        localtrack_folder = aiopath.AsyncPath(folder) if folder else self._config_folder / "music"

        await LocalFile.add_root_folder(path=localtrack_folder, create=True)

    async def get_all_players(self) -> AsyncIterator[Player]:

        return asyncstdlib.iter(self.player_manager)

    async def get_managed_node(self) -> Node | None:
        available_nodes = await asyncstdlib.list(
            asyncstdlib.filter(operator.attrgetter("available"), self.node_manager.managed_nodes)
        )

        return random.choice(available_nodes) if available_nodes else None

    async def _get_tracks(
        self,
        query: Query,
        first: bool = False,
        bypass_cache: bool = False,
        player: Player | None = None,
    ) -> dict:
        """|coro|
        Gets all tracks associated with the given query.

        Parameters
        ----------
        query: :class:`Query`
            The query to perform a search for
        first: Optional[:class:`bool`]
            Whether to only return the first track. Defaults to `False`.
        bypass_cache: Optional[:class:`bool`]
            Whether to bypass the cache. Defaults to `False`.

        Returns
        -------
        :class:`dict`
            A dict representing tracks.
        """
        if not self.node_manager.available_nodes:
            raise NoNodeAvailable("No available nodes!")
        node = await self.node_manager.find_best_node(
            feature=query.requires_capability,
            region=player.region if player else None,
            coordinates=player.coordinates if player else None,
        )
        if node is None:
            raise NoNodeWithRequestFunctionalityAvailable(
                _("No node with {query.requires_capability} functionality available!").format(query=query),
                query.requires_capability,
            )
        return await node.get_tracks(query, first=first, bypass_cache=bypass_cache)

    async def get_all_tracks_for_queries(
        self,
        *queries: Query,
        requester: discord.Member,
        player: Player | None = None,
        bypass_cache: bool = False,
        enqueue: bool = True,
    ) -> tuple[list[Track], int, list[Query]]:  # sourcery no-metrics
        """High level interface to get and return all tracks for a list of queries.

        This will automatically handle playlists, albums, searches and local files.

        Parameters
        ----------
        queries : `Query`
            The list of queries to search for.
        bypass_cache : `bool`, optional
            Whether to bypass the cache and force a new search.
            Local files will always be bypassed.
        requester : `discord.Nember`
            The user who requested the op.
        player : `Player`
            The player requesting the op.
        enqueue : `bool`, optional
            Whether to enqueue the tracks as needed
            while try are processed so users dont sit waiting for the bot to finish.

        Returns
        -------
        tracks : `List[Track]`
            The list of tracks found.
        total_tracks : `int`
            The total number of tracks found.
        queries : `List[Query]`
            The list of queries that were not found.

        """
        successful_tracks = []
        queries_failed = []
        track_count = 0

        for query in queries:
            node = await self.node_manager.find_best_node(
                region=player.region if player else None,
                coordinates=player.coordinates if player else None,
                feature=query.requires_capability,
            )
            if node is None:
                queries_failed.append(query)
            # Query tracks as the queue builds as this may be a slow operation
            if enqueue and successful_tracks and not player.is_playing and not player.paused:
                track = successful_tracks.pop()
                await player.play(track, await track.query(), requester)
            if query.is_search or query.is_single:
                try:
                    track = await self._get_tracks(player=player, query=query, first=True, bypass_cache=bypass_cache)
                    track_b64 = track.get("track")
                    if not track_b64:
                        queries_failed.append(query)
                    if track_b64:
                        track_count += 1
                        successful_tracks.append(
                            Track(
                                data=track_b64,
                                node=node,
                                query=await Query.from_base64(track_b64),
                                requester=requester.id,
                            )
                        )
                except NoNodeWithRequestFunctionalityAvailable:
                    queries_failed.append(query)
            elif (query.is_playlist or query.is_album) and not query.is_local and not query.is_custom_playlist:
                try:
                    tracks: dict = await self._get_tracks(player=player, query=query, bypass_cache=bypass_cache)
                    track_list = tracks.get("tracks", [])
                    if not track_list:
                        queries_failed.append(query)
                    for track in track_list:
                        if track_b64 := track.get("track"):
                            track_count += 1
                            successful_tracks.append(
                                Track(
                                    data=track_b64,
                                    node=node,
                                    query=await Query.from_base64(track_b64),
                                    requester=requester.id,
                                )
                            )
                            # Query tracks as the queue builds as this may be a slow operation
                            if enqueue and successful_tracks and not player.is_playing and not player.paused:
                                track = successful_tracks.pop()
                                await player.play(track, await track.query(), requester)
                except NoNodeWithRequestFunctionalityAvailable:
                    queries_failed.append(query)
            elif (query.is_local or query.is_custom_playlist) and query.is_album:
                try:
                    yielded = False
                    async for local_track in query.get_all_tracks_in_folder():
                        yielded = True
                        track = await self._get_tracks(player=player, query=local_track, first=True, bypass_cache=True)
                        if track_b64 := track.get("track"):
                            track_count += 1
                            successful_tracks.append(
                                Track(
                                    data=track_b64,
                                    node=node,
                                    query=await Query.from_base64(track_b64),
                                    requester=requester.id,
                                )
                            )
                            # Query tracks as the queue builds as this may be a slow operation
                            if enqueue and successful_tracks and not player.is_playing and not player.paused:
                                track = successful_tracks.pop()
                                await player.play(track, await track.query(), requester)
                    if not yielded:
                        queries_failed.append(query)
                except NoNodeWithRequestFunctionalityAvailable:
                    queries_failed.append(query)
            else:
                queries_failed.append(query)
                LOGGER.warning("Unhandled query: %s, %s", query.__dict__, query.query_identifier)
        return successful_tracks, track_count, queries_failed

    @staticmethod
    async def _yield_recursive_queries(query: Query, recursion_depth: int = 0) -> AsyncIterator[Query]:
        """|coro|
        Gets all queries associated with the given query.
        Parameters
        ----------
        query: :class:`Query`
            The query to perform a search for
        """
        if query.invalid or recursion_depth > MAX_RECURSION_DEPTH:
            return
        recursion_depth += 1
        if query.is_m3u:
            async for m3u in query._yield_m3u_tracks():
                with contextlib.suppress(Exception):
                    async for q in query._yield_tracks_recursively(m3u, recursion_depth):
                        yield q
        elif query.is_pylav:
            async for pylav in query._yield_pylav_file_tracks():
                with contextlib.suppress(Exception):
                    async for q in query._yield_tracks_recursively(pylav, recursion_depth):
                        yield q
        elif query.is_pls:
            async for pls in query._yield_pls_tracks():
                with contextlib.suppress(Exception):
                    async for q in query._yield_tracks_recursively(pls, recursion_depth):
                        yield q
        elif query.is_local and query.is_album:
            async for local in query._yield_local_tracks():
                yield local
        else:
            yield query

    async def get_tracks(
        self,
        *queries: Query,
        bypass_cache: bool = False,
        fullsearch: bool = False,
        region: str | None = None,
        player: Player | None = None,
    ) -> LavalinkResponseT:
        """This method can be rather slow as it recursibly queries all queries and their associated entries.

        Thus if you are processing user input  you may be interested in using
        the :meth:`get_all_tracks_for_queries` where it can enqueue tracks as needed to the player.


        Parameters
        ----------
        queries : `Query`
            The list of queries to search for.
        bypass_cache : `bool`, optional
            Whether to bypass the cache and force a new search.
            Local files will always be bypassed.
        fullsearch : `bool`, optional
            if a Search query is passed wether to returrn a list of tracks instead of the first.
        region : `str`, optional
            The region to search in.
        player : `Player`, optional
            The player to use for enqueuing tracks.
        """
        output_tracks = []
        playlist_name = ""
        if region is None and player is not None:
            region = player.region

        if region is None:
            region = "us_east"

        for query in queries:
            async for response in self._yield_recursive_queries(query):
                with contextlib.suppress(NoNodeWithRequestFunctionalityAvailable):
                    node = await self.node_manager.find_best_node(region=region, feature=response.requires_capability)
                    if node is None or response.is_custom_playlist:
                        continue
                    if response.is_playlist or response.is_album:
                        _response = await node.get_tracks(response, bypass_cache=bypass_cache)
                        playlist_name = _response.get("playlistInfo", {}).get("name", "")
                        output_tracks.extend(_response["tracks"])
                    elif fullsearch and response.is_search:
                        _response = await node.get_tracks(response, bypass_cache=bypass_cache)
                        output_tracks.extend(_response["tracks"])
                    elif response.is_single:
                        _response = await node.get_tracks(response, first=True, bypass_cache=bypass_cache)
                        output_tracks.append(_response)
                    else:
                        LOGGER.critical("Unknown query type: %s", response)

        return {
            "playlistInfo": {
                "name": playlist_name if len(queries) == 1 else "",
                "selectedTrack": -1,
            },
            "loadType": "PLAYLIST_LOADED" if playlist_name else "SEARCH_RESULT" if output_tracks else "LOAD_FAILED",  # type: ignore
            "tracks": output_tracks,
        }

    async def remove_node(self, node_id: int):
        """Removes a node from the node manager."""
        node = self.node_manager.get_node_by_id(node_id)
        await self.node_manager.remove_node(node)

    async def is_dj(
        self,
        user: discord.Member,
        guild: discord.Guild,
        *,
        additional_role_ids: list = None,
        additional_user_ids: list = None,
        bot: BotT = None,
    ) -> bool:
        if additional_user_ids and user.id in additional_user_ids:
            return True
        if additional_role_ids and await asyncstdlib.any(r.id in additional_role_ids for r in user.roles):
            return True
        return await self.player_config_manager.is_dj(
            user=user, guild=guild, additional_role_ids=None, additional_user_ids=None, bot=bot
        )

    async def generate_mix_playlist(
        self,
        *,
        video_id: str | None = None,
        user_id: str | None,
        playlist_id: str | None = None,
        channel_id: str | None = None,
    ) -> str:
        if not await asyncstdlib.any([video_id, playlist_id, channel_id, user_id]):
            raise PyLavInvalidArguments(
                _("To generate a mix playlist a Video, User, Channel or Playlist ID is necessary.")
            )

        if await asyncstdlib.sum(1 for i in [video_id, playlist_id, channel_id, user_id] if i) > 1:
            raise PyLavInvalidArguments(
                _(
                    "To generate a mix playlist a Video, User, Channel or Playlist ID is necessary. "
                    "However, you provided multiple."
                )
            )

        if video_id:
            return f"https://www.youtube.com/watch?list=RD{video_id}"
        if user_id:
            return f"https://www.youtube.com/watch?list=UU{user_id}"
        if playlist_id:
            return f"https://www.youtube.com/watch?list=RDAMPL{playlist_id}"
        if channel_id:
            return f"https://www.youtube.com/watch?list=RDCM{channel_id}"



from __future__ import annotations

import contextlib
from typing import TYPE_CHECKING

from packaging.version import parse as parse_version

if TYPE_CHECKING:
    from pylav.client import Client


class UpdateSchemaManager:
    def __init__(self, client: Client):
        self._client = client

    async def run_updates(self):
        """Run through schema migrations."""
        from pylav import EntryNotFoundError
        from pylav._config import __VERSION__

        # FIXME: This should be whatever value the first release is or alternatively `self._client.lib_version`

        if (await self._client.lib_db_manager.get_bot_db_version()).version <= parse_version("0.0.0.0"):
            await self._client.lib_db_manager.update_bot_dv_version("0.0.0.1")

        if (await self._client.lib_db_manager.get_bot_db_version()).version <= parse_version("0.0.0.1"):
            full_data = await self._client.node_db_manager.get_bundled_node_config()
            full_data.yaml["lavalink"]["server"]["trackStuckThresholdMs"] = 10000
            await full_data.save()
            await self._client.lib_db_manager.update_bot_dv_version("0.0.0.2")

        if (await self._client.lib_db_manager.get_bot_db_version()).version <= parse_version("0.3.1"):
            full_data = await self._client.node_db_manager.get_bundled_node_config()
            full_data.yaml["lavalink"]["server"]["opusEncodingQuality"] = 10
            full_data.yaml["lavalink"]["server"]["resamplingQuality"] = "LOW"
            full_data.yaml["lavalink"]["server"]["useSeekGhosting"] = True
            await full_data.save()
            await self._client.lib_db_manager.update_bot_dv_version("0.3.2")

        if (await self._client.lib_db_manager.get_bot_db_version()).version <= parse_version("0.3.2"):
            full_data = await self._client.node_db_manager.get_bundled_node_config()
            full_data.yaml["lavalink"]["server"]["youtubeConfig"] = {"email": "", "password": ""}
            await full_data.save()
            await self._client.lib_db_manager.update_bot_dv_version("0.3.3")

        if (await self._client.lib_db_manager.get_bot_db_version()).version <= parse_version("0.3.3"):
            full_data = await self._client.node_db_manager.get_bundled_node_config()
            if "soundgasm" not in full_data.yaml["plugins"]["dunctebot"]["sources"]:
                full_data.yaml["plugins"]["dunctebot"]["sources"]["soundgasm"] = True

            full_data.yaml["lavalink"]["plugins"] = [
                {
                    "dependency": "com.github.Topis-Lavalink-Plugins:Topis-Source-Managers-Plugin:v2.0.7",
                    "repository": "https://jitpack.io",
                },
                {
                    "dependency": "com.dunctebot:skybot-lavalink-plugin:1.4.0",
                    "repository": "https://m2.duncte123.dev/releases",
                },
                {"dependency": "com.github.topisenpai:sponsorblock-plugin:v1.0.3", "repository": "https://jitpack.io"},
            ]
            await full_data.save()
            await self._client.lib_db_manager.update_bot_dv_version("0.3.4")

        if (await self._client.lib_db_manager.get_bot_db_version()).version <= parse_version("0.3.4"):
            full_data = await self._client.node_db_manager.get_bundled_node_config()
            if "path" in full_data.yaml["logging"]:
                full_data.yaml["logging"]["file"]["path"] = full_data.yaml["logging"]["path"]
                await full_data.save()
            await self._client.lib_db_manager.update_bot_dv_version("0.3.5")

        if (await self._client.lib_db_manager.get_bot_db_version()).version <= parse_version("0.3.5"):
            full_data = await self._client.node_db_manager.get_bundled_node_config()
            if "path" in full_data.yaml["logging"]:
                del full_data.yaml["logging"]["path"]
            if "rollingpolicy" not in full_data.yaml["logging"]["logback"]:
                full_data.yaml["logging"]["logback"] = {
                    "rollingpolicy": {
                        "max-file-size": full_data.yaml["logging"]["file"]["max-size"],
                        "max-history": full_data.yaml["logging"]["file"]["max-history"],
                        "total-size-cap": "1GB",
                    }
                }
            if "max-size" in full_data.yaml["logging"]["file"]:
                del full_data.yaml["logging"]["file"]["max-size"]
            if "max-history" in full_data.yaml["logging"]["file"]:
                del full_data.yaml["logging"]["file"]["max-history"]
            await full_data.save()
            await self._client.lib_db_manager.update_bot_dv_version("0.3.6")
        with contextlib.suppress(EntryNotFoundError):
            bundled_external_node = await self._client.node_db_manager.get_node_config(1)
            bundled_external_node.resume_key = f"PyLav/{self._client.lib_version}-{self._client.bot_id}"
            await bundled_external_node.save()
        await self._client.lib_db_manager.update_bot_dv_version(__VERSION__)



from __future__ import annotations

from collections.abc import AsyncIterator
from typing import TYPE_CHECKING

import discord

from pylav._logging import getLogger
from pylav.exceptions import EntryNotFoundError
from pylav.sql import tables
from pylav.sql.models import EqualizerModel
from pylav.types import BotT
from pylav.utils import AsyncIter

if TYPE_CHECKING:
    from pylav.client import Client

LOGGER = getLogger("PyLav.EqualizerConfigManager")


class EqualizerConfigManager:
    def __init__(self, client: Client):
        self._client = client

    @property
    def client(self) -> Client:
        return self._client

    @staticmethod
    async def get_equalizer_by_name(equalizer_name: str, limit: int = None) -> list[EqualizerModel]:
        if limit is None:
            equalizers = await tables.EqualizerRow.select().where(
                tables.EqualizerRow.name.ilike(f"%{equalizer_name.lower()}%")
            )
        else:
            equalizers = (
                await tables.EqualizerRow.select()
                .where(tables.EqualizerRow.name.ilike(f"%{equalizer_name.lower()}%"))
                .limit(limit)
            )

        if not equalizers:
            raise EntryNotFoundError(f"Equalizer with name {equalizer_name} not found")
        return [EqualizerModel(**equalizer) for equalizer in equalizers]

    @staticmethod
    async def get_equalizer_by_id(equalizer_id: int | str) -> EqualizerModel:
        try:
            equalizer_id = int(equalizer_id)
        except ValueError as e:
            raise EntryNotFoundError(f"Equalizer with id {equalizer_id} not found") from e
        equalizer = await tables.EqualizerRow.select().where(tables.EqualizerRow.id == equalizer_id).limit(1).first()
        if not equalizer:
            raise EntryNotFoundError(f"Equalizer with ID {equalizer_id} not found")
        return EqualizerModel(**equalizer)

    async def get_equalizer_by_name_or_id(
        self, equalizer_name_or_id: int | str, limit: int = None
    ) -> list[EqualizerModel]:
        try:
            return [await self.get_equalizer_by_id(equalizer_name_or_id)]
        except EntryNotFoundError:
            return await self.get_equalizer_by_name(equalizer_name_or_id, limit=limit)

    @staticmethod
    async def get_equalizers_by_author(author: int) -> list[EqualizerModel]:
        equalizers = await tables.EqualizerRow.select().where(tables.EqualizerRow.author == author)
        if not equalizers:
            raise EntryNotFoundError(f"Equalizer with author {author} not found")
        return [EqualizerModel(**equalizer) for equalizer in equalizers]

    @staticmethod
    async def get_equalizers_by_scope(scope: int) -> list[EqualizerModel]:
        equalizers = await tables.EqualizerRow.select().where(tables.EqualizerRow.scope == scope)
        if not equalizers:
            raise EntryNotFoundError(f"Equalizer with scope {scope} not found")
        return [EqualizerModel(**equalizer) for equalizer in equalizers]

    @staticmethod
    async def get_all_equalizers() -> AsyncIterator[EqualizerModel]:
        for entry in await tables.EqualizerRow.select():
            yield EqualizerModel(**entry)

    @staticmethod
    async def create_or_update_equalizer(
        id: int,
        scope: int,
        author: int,
        name: str,
        description: str | None = None,
        band_25: float = 0.0,
        band_40: float = 0.0,
        band_63: float = 0.0,
        band_100: float = 0.0,
        band_160: float = 0.0,
        band_250: float = 0.0,
        band_400: float = 0.0,
        band_630: float = 0.0,
        band_1000: float = 0.0,
        band_1600: float = 0.0,
        band_2500: float = 0.0,
        band_4000: float = 0.0,
        band_6300: float = 0.0,
        band_10000: float = 0.0,
        band_16000: float = 0.0,
    ) -> EqualizerModel:
        values = {
            tables.EqualizerRow.scope: scope,
            tables.EqualizerRow.author: author,
            tables.EqualizerRow.name: name,
            tables.EqualizerRow.description: description,
            tables.EqualizerRow.band_25: band_25,
            tables.EqualizerRow.band_40: band_40,
            tables.EqualizerRow.band_63: band_63,
            tables.EqualizerRow.band_100: band_100,
            tables.EqualizerRow.band_160: band_160,
            tables.EqualizerRow.band_250: band_250,
            tables.EqualizerRow.band_400: band_400,
            tables.EqualizerRow.band_630: band_630,
            tables.EqualizerRow.band_1000: band_1000,
            tables.EqualizerRow.band_1600: band_1600,
            tables.EqualizerRow.band_2500: band_2500,
            tables.EqualizerRow.band_4000: band_4000,
            tables.EqualizerRow.band_6300: band_6300,
            tables.EqualizerRow.band_10000: band_10000,
            tables.EqualizerRow.band_16000: band_16000,
        }
        equalizer = (
            await tables.EqualizerRow.objects()
            .output(load_json=True)
            .get_or_create(tables.EqualizerRow.id == id, defaults=values)
        )
        if not equalizer._was_created:
            await tables.EqualizerRow.update(values).where(tables.EqualizerRow.id == id)
        return EqualizerModel(**equalizer.to_dict())

    @staticmethod
    async def delete_equalizer(equalizer_id: int) -> None:
        await tables.EqualizerRow.delete().where(tables.EqualizerRow.id == equalizer_id)

    @staticmethod
    async def get_all_equalizers_by_author(author: int) -> AsyncIterator[EqualizerModel]:
        for entry in await tables.EqualizerRow.select().where(tables.EqualizerRow.author == author):
            yield EqualizerModel(**entry)

    @staticmethod
    async def get_all_equalizers_by_scope(scope: int) -> AsyncIterator[EqualizerModel]:
        for entry in await tables.EqualizerRow.select().where(tables.EqualizerRow.scope == scope):
            yield EqualizerModel(**entry)

    @staticmethod
    async def get_all_equalizers_by_scope_and_author(scope: int, author: int) -> AsyncIterator[EqualizerModel]:
        for entry in await tables.EqualizerRow.select().where(
            tables.EqualizerRow.scope == scope, tables.EqualizerRow.author == author
        ):
            yield EqualizerModel(**entry)

    async def get_global_equalizers(self) -> AsyncIterator[EqualizerModel]:
        for entry in await tables.EqualizerRow.select().where(tables.EqualizerRow.scope == self._client.bot.user.id):  # type: ignore
            yield EqualizerModel(**entry)

    async def create_or_update_global_equalizer(
        self,
        id: int,
        author: int,
        name: str,
        description: str | None = None,
        band_25: float = 0.0,
        band_40: float = 0.0,
        band_63: float = 0.0,
        band_100: float = 0.0,
        band_160: float = 0.0,
        band_250: float = 0.0,
        band_400: float = 0.0,
        band_630: float = 0.0,
        band_1000: float = 0.0,
        band_1600: float = 0.0,
        band_2500: float = 0.0,
        band_4000: float = 0.0,
        band_6300: float = 0.0,
        band_10000: float = 0.0,
        band_16000: float = 0.0,
    ) -> EqualizerModel:
        return await self.create_or_update_equalizer(
            id=id,
            scope=self._client.bot.user.id,
            author=author,
            name=name,
            description=description,
            band_25=band_25,
            band_40=band_40,
            band_63=band_63,
            band_100=band_100,
            band_160=band_160,
            band_250=band_250,
            band_400=band_400,
            band_630=band_630,
            band_1000=band_1000,
            band_1600=band_1600,
            band_2500=band_2500,
            band_4000=band_4000,
            band_6300=band_6300,
            band_10000=band_10000,
            band_16000=band_16000,
        )

    async def create_or_update_user_equalizer(
        self,
        id: int,
        author: int,
        name: str,
        description: str | None = None,
        band_25: float = 0.0,
        band_40: float = 0.0,
        band_63: float = 0.0,
        band_100: float = 0.0,
        band_160: float = 0.0,
        band_250: float = 0.0,
        band_400: float = 0.0,
        band_630: float = 0.0,
        band_1000: float = 0.0,
        band_1600: float = 0.0,
        band_2500: float = 0.0,
        band_4000: float = 0.0,
        band_6300: float = 0.0,
        band_10000: float = 0.0,
        band_16000: float = 0.0,
    ) -> EqualizerModel:
        return await self.create_or_update_equalizer(
            id=id,
            scope=author,
            author=author,
            name=name,
            description=description,
            band_25=band_25,
            band_40=band_40,
            band_63=band_63,
            band_100=band_100,
            band_160=band_160,
            band_250=band_250,
            band_400=band_400,
            band_630=band_630,
            band_1000=band_1000,
            band_1600=band_1600,
            band_2500=band_2500,
            band_4000=band_4000,
            band_6300=band_6300,
            band_10000=band_10000,
            band_16000=band_16000,
        )

    async def create_or_update_channel_equalizer(
        self,
        channel: discord.abc.MessageableChannel,
        id: int,
        author: int,
        name: str,
        description: str | None = None,
        band_25: float = 0.0,
        band_40: float = 0.0,
        band_63: float = 0.0,
        band_100: float = 0.0,
        band_160: float = 0.0,
        band_250: float = 0.0,
        band_400: float = 0.0,
        band_630: float = 0.0,
        band_1000: float = 0.0,
        band_1600: float = 0.0,
        band_2500: float = 0.0,
        band_4000: float = 0.0,
        band_6300: float = 0.0,
        band_10000: float = 0.0,
        band_16000: float = 0.0,
    ) -> EqualizerModel:
        return await self.create_or_update_equalizer(
            id=id,
            scope=channel.id,
            author=author,
            name=name,
            description=description,
            band_25=band_25,
            band_40=band_40,
            band_63=band_63,
            band_100=band_100,
            band_160=band_160,
            band_250=band_250,
            band_400=band_400,
            band_630=band_630,
            band_1000=band_1000,
            band_1600=band_1600,
            band_2500=band_2500,
            band_4000=band_4000,
            band_6300=band_6300,
            band_10000=band_10000,
            band_16000=band_16000,
        )

    async def create_or_update_guild_equalizer(
        self,
        guild: discord.Guild,
        id: int,
        author: int,
        name: str,
        description: str | None = None,
        band_25: float = 0.0,
        band_40: float = 0.0,
        band_63: float = 0.0,
        band_100: float = 0.0,
        band_160: float = 0.0,
        band_250: float = 0.0,
        band_400: float = 0.0,
        band_630: float = 0.0,
        band_1000: float = 0.0,
        band_1600: float = 0.0,
        band_2500: float = 0.0,
        band_4000: float = 0.0,
        band_6300: float = 0.0,
        band_10000: float = 0.0,
        band_16000: float = 0.0,
    ) -> EqualizerModel:
        return await self.create_or_update_equalizer(
            id=id,
            scope=guild.id,
            author=author,
            name=name,
            description=description,
            band_25=band_25,
            band_40=band_40,
            band_63=band_63,
            band_100=band_100,
            band_160=band_160,
            band_250=band_250,
            band_400=band_400,
            band_630=band_630,
            band_1000=band_1000,
            band_1600=band_1600,
            band_2500=band_2500,
            band_4000=band_4000,
            band_6300=band_6300,
            band_10000=band_10000,
            band_16000=band_16000,
        )

    async def create_or_update_vc_equalizer(
        self,
        vc: discord.channel.VocalGuildChannel,
        id: int,
        author: int,
        name: str,
        description: str | None = None,
        band_25: float = 0.0,
        band_40: float = 0.0,
        band_63: float = 0.0,
        band_100: float = 0.0,
        band_160: float = 0.0,
        band_250: float = 0.0,
        band_400: float = 0.0,
        band_630: float = 0.0,
        band_1000: float = 0.0,
        band_1600: float = 0.0,
        band_2500: float = 0.0,
        band_4000: float = 0.0,
        band_6300: float = 0.0,
        band_10000: float = 0.0,
        band_16000: float = 0.0,
    ) -> EqualizerModel:
        return await self.create_or_update_equalizer(
            id=id,
            scope=vc.id,
            author=author,
            name=name,
            description=description,
            band_25=band_25,
            band_40=band_40,
            band_63=band_63,
            band_100=band_100,
            band_160=band_160,
            band_250=band_250,
            band_400=band_400,
            band_630=band_630,
            band_1000=band_1000,
            band_1600=band_1600,
            band_2500=band_2500,
            band_4000=band_4000,
            band_6300=band_6300,
            band_10000=band_10000,
            band_16000=band_16000,
        )

    async def get_all_for_user(
        self,
        requester: int,
        *,
        vc: discord.channel.VocalGuildChannel = None,
        guild: discord.Guild = None,
        channel: discord.abc.MessageableChannel = None,
    ) -> tuple[
        list[EqualizerModel], list[EqualizerModel], list[EqualizerModel], list[EqualizerModel], list[EqualizerModel]
    ]:
        """
        Gets all equalizers a user has access to in a given context.

        Globals, User specific, Guild specific, Channel specific, VC specific.

        """
        global_equalizers = [p async for p in self.get_all_equalizers_by_scope(scope=self._client.bot.user.id)]
        user_equalizers = [p async for p in self.get_all_equalizers_by_scope(scope=requester)]
        vc_equalizers = []
        guild_equalizers = []
        channel_equalizers = []
        if vc is not None:
            vc_equalizers = [p async for p in self.get_all_equalizers_by_scope(scope=vc.id)]
        if guild is not None:
            guild_equalizers = [p async for p in self.get_all_equalizers_by_scope(scope=guild.id)]
        if channel is not None:
            channel_equalizers = [p async for p in self.get_all_equalizers_by_scope(scope=channel.id)]
        return global_equalizers, user_equalizers, guild_equalizers, channel_equalizers, vc_equalizers

    async def get_manageable_equalizers(
        self, requester: discord.abc.User, bot: BotT, *, name_or_id: str | None = None
    ) -> list[EqualizerModel]:
        if name_or_id:
            try:
                equalizers = await self.get_equalizer_by_name_or_id(name_or_id)
            except EntryNotFoundError:
                equalizers = []
        else:
            try:
                equalizers = [p async for p in self.get_all_equalizers()]
            except EntryNotFoundError:
                equalizers = []
        returning_list = []
        if equalizers:
            async for equalizer in AsyncIter(equalizers):
                if await equalizer.can_manage(requester=requester, bot=bot):
                    returning_list.append(equalizer)
        return returning_list



# sourcery skip: avoid-builtin-shadow
from __future__ import annotations

import os

from piccolo.columns import JSONB, UUID, Array, BigInt, Boolean, Bytea, Float, Integer, Text, Timestamptz
from piccolo.engine import PostgresEngine, SQLiteEngine
from piccolo.table import Table

from pylav._logging import getLogger
from pylav.envvars import (
    JAVA_EXECUTABLE,
    POSTGRES_DATABASE,
    POSTGRES_HOST,
    POSTGRES_PASSWORD,
    POSTGRES_PORT,
    POSTGRES_USER,
)

LOGGER = getLogger("PyLav.Postgres")
config = {
    "host": POSTGRES_HOST,
    "port": int(POSTGRES_PORT) if isinstance(POSTGRES_PORT, str) else POSTGRES_PORT,
    "database": POSTGRES_DATABASE,
    "user": POSTGRES_USER,
    "password": POSTGRES_PASSWORD,
}
LOGGER.info("Connecting to Postgres server using %r", config)
if os.getenv("BUILDING_DOCS", False):
    DB = SQLiteEngine()
else:
    DB = PostgresEngine(config=config)


class PlaylistRow(Table, db=DB, tablename="playlist"):
    id = BigInt(primary_key=True)
    scope = BigInt(null=False, index=True)
    author = BigInt(null=False, index=True)
    name = Text(null=False, index=True)
    url = Text(null=True)
    tracks = JSONB(null=False, default=[])


class LibConfigRow(Table, db=DB, tablename="lib_config"):
    bot = BigInt(primary_key=True, index=True)
    id = BigInt(index=True)
    config_folder = Text(null=False)
    localtrack_folder = Text(null=True)
    java_path = Text(null=False, default=JAVA_EXECUTABLE)
    enable_managed_node = Boolean(null=False, default=True)
    use_bundled_lava_link_external = Boolean(null=False, default=True)
    use_bundled_pylav_external = Boolean(null=False, default=True)
    auto_update_managed_nodes = Boolean(null=False, default=True)
    download_id = BigInt(index=True, default=0)
    extras: JSONB = JSONB(null=False, default={})
    next_execution_update_bundled_playlists = Timestamptz(null=True)
    next_execution_update_bundled_external_playlists = Timestamptz(null=True)
    next_execution_update_external_playlists = Timestamptz(null=True)
    update_bot_activity = Boolean(null=False, default=False)


class EqualizerRow(Table, db=DB, tablename="equalizer"):
    id = BigInt(primary_key=True, index=True)
    scope = BigInt(null=False, index=True)
    name = Text(null=True, index=True)
    description = Text(null=True)
    author = BigInt(null=False, index=True)
    band_25 = Float(null=False, default=0.0)
    band_40 = Float(null=False, default=0.0)
    band_63 = Float(null=False, default=0.0)
    band_100 = Float(null=False, default=0.0)
    band_160 = Float(null=False, default=0.0)
    band_250 = Float(null=False, default=0.0)
    band_400 = Float(null=False, default=0.0)
    band_630 = Float(null=False, default=0.0)
    band_1000 = Float(null=False, default=0.0)
    band_1600 = Float(null=False, default=0.0)
    band_2500 = Float(null=False, default=0.0)
    band_4000 = Float(null=False, default=0.0)
    band_6300 = Float(null=False, default=0.0)
    band_10000 = Float(null=False, default=0.0)
    band_16000 = Float(null=False, default=0.0)


class PlayerStateRow(Table, db=DB, tablename="player_state"):
    pk = UUID(primary_key=True)
    id = BigInt(index=True)
    bot = BigInt(index=True, null=False)
    channel_id = BigInt(null=False)
    volume = Integer(null=False, default=100)
    position = Float(null=False, default=0.0)
    auto_play_playlist_id = BigInt(null=True)
    forced_channel_id = BigInt(null=True)
    text_channel_id = BigInt(null=True)
    notify_channel_id = BigInt(null=True)

    paused = Boolean(null=False, default=False)
    repeat_current = Boolean(null=False, default=False)
    repeat_queue = Boolean(null=False, default=False)
    shuffle = Boolean(null=False, default=True)
    auto_shuffle = Boolean(null=False, default=False)
    auto_play = Boolean(null=False, default=False)
    playing = Boolean(null=False, default=False)
    effect_enabled = Boolean(null=False, default=False)
    self_deaf = Boolean(null=False, default=False)

    current = JSONB(null=True)
    queue = JSONB(null=False, default=[])
    history = JSONB(null=False, default=[])
    effects = JSONB(null=False, default={})
    extras = JSONB(null=False, default={})


class PlayerRow(Table, db=DB, tablename="player"):
    pk = UUID(primary_key=True)
    id = BigInt(index=True)
    bot = BigInt(index=True, null=False)
    volume = Integer(null=False, default=100)
    max_volume = Integer(null=False, default=1000)
    auto_play_playlist_id = BigInt(null=False, default=1)

    text_channel_id = BigInt(null=True)
    notify_channel_id = BigInt(null=True)
    forced_channel_id = BigInt(null=True)

    repeat_current = Boolean(null=False, default=False)
    repeat_queue = Boolean(null=False, default=False)
    shuffle = Boolean(null=False, default=True)
    auto_shuffle = Boolean(null=False, default=True)
    auto_play = Boolean(null=False, default=True)
    self_deaf = Boolean(null=False, default=True)
    empty_queue_dc = JSONB(
        null=False,
        default={
            "enabled": False,
            "time": 60,
        },
    )
    alone_dc = JSONB(
        null=False,
        default={
            "enabled": False,
            "time": 60,
        },
    )
    alone_pause = JSONB(
        null=False,
        default={
            "enabled": False,
            "time": 60,
        },
    )
    extras = JSONB(null=False, default={})
    effects = JSONB(null=False, default={})
    dj_users = Array(null=False, default=[], base_column=BigInt())
    dj_roles = Array(null=False, default=[], base_column=BigInt())


class NodeRow(Table, db=DB, tablename="node"):
    id = BigInt(primary_key=True, index=True)
    name = Text(null=False)
    ssl = Boolean(null=False, default=False)
    resume_key = Text(null=True)
    resume_timeout = Integer(null=False, default=600)
    reconnect_attempts = Integer(null=False)
    search_only = Boolean(null=False, default=False)
    managed = Boolean(null=False, default=False)
    disabled_sources = JSONB(null=False, default=[])
    extras = JSONB(null=True)
    yaml = JSONB(null=True)


class QueryRow(Table, db=DB, tablename="query"):
    identifier = Text(null=False, index=True, primary_key=True)
    name = Text(null=True)
    tracks = JSONB(null=False, default=[])
    last_updated = Timestamptz(null=False, index=True)


class BotVersionRow(Table, db=DB, tablename="version"):
    bot = BigInt(primary_key=True, index=True)
    version = Text(null=False)


class AioHttpCacheRow(Table, db=DB, tablename="aiohttp_client_cache"):
    key = Text(primary_key=True, index=True)
    value = Bytea()



from __future__ import annotations

from collections.abc import AsyncIterator
from typing import TYPE_CHECKING

from pylav._logging import getLogger
from pylav.sql import tables
from pylav.sql.models import PlayerStateModel

if TYPE_CHECKING:
    from pylav.client import Client

LOGGER = getLogger("PyLav.PlayerStateDBManager")


class PlayerStateDBManager:
    def __init__(self, client: Client):
        self._client = client

    @property
    def client(self) -> Client:
        return self._client

    async def save_players(self, players: list[dict]):
        for player in players:
            p = PlayerStateModel(bot=self.client.bot.user.id, **player)
            await p.save()
            LOGGER.trace("Saved player %s", p)
        LOGGER.debug("Saved %s players", len(players))

    async def save_player(self, player: dict):
        await PlayerStateModel(bot=self.client.bot.user.id, **player).save()

    async def get_player(self, guild_id: int) -> PlayerStateModel | None:
        return await PlayerStateModel.get(bot_id=self._client.bot.user.id, guild_id=guild_id)

    async def get_all_players(self) -> AsyncIterator[PlayerStateModel]:
        for entry in await tables.PlayerStateRow.select().where(tables.PlayerStateRow.bot == self.client.bot.user.id):  # type: ignore
            yield PlayerStateModel(**entry)

    async def delete_player(self, guild_id: int):
        await tables.PlayerStateRow.delete().where(
            (tables.PlayerStateRow.bot == self.client.bot.user.id) & (tables.PlayerStateRow.id == guild_id)
        )



from __future__ import annotations

import aiohttp
from discord.ext.commands import CommandError

__all__ = (
    "PyLavError",
    "NodeError",
    "WebsocketNotConnectedError",
    "TrackError",
    "ManagedLavalinkNodeError",
    "HTTPError",
    "Unauthorized",
    "InvalidTrack",
    "NodeUnhealthy",
    "InvalidArchitectureError",
    "ManagedLavalinkAlreadyRunningError",
    "PortAlreadyInUseError",
    "ManagedLavalinkStartFailure",
    "ManagedLavalinkPreviouslyShutdownError",
    "EarlyExitError",
    "UnsupportedJavaError",
    "UnexpectedJavaResponseError",
    "NoProcessFound",
    "IncorrectProcessFound",
    "TooManyProcessFound",
    "LavalinkDownloadFailed",
    "TrackNotFound",
    "CogAlreadyRegistered",
    "CogHasBeenRegistered",
    "AnotherClientAlreadyRegistered",
    "NoNodeAvailable",
    "SQLError",
    "EntryNotFoundError",
    "PyLavNotInitialized",
    "PlaylistError",
    "InvalidPlaylist",
    "ManagedLinkStartAbortedUseExternal",
    "AbortPlayerRestoreDueUnavailableNode",
    "NoNodeWithRequestFunctionalityAvailable",
)


class PyLavError(CommandError):
    """Base exception for errors in the library."""


class PyLavInvalidArguments(PyLavError):
    """Base Exception for when invalid arguments are passed to a method"""


class PyLavNotInitialized(PyLavError):
    """Raised when the library is not initialized."""


class SQLError(PyLavError):
    """Base exception for errors in SQL."""


class EntryNotFoundError(SQLError):
    """Raised when an entry is not found."""


class AnotherClientAlreadyRegistered(PyLavError):
    """Another client has already been registered."""


class NodeError(PyLavError):
    """Base exception for Node errors."""


class AbortPlayerRestoreDueUnavailableNode(NodeError):
    """Raised when the player is aborted due to an unavailable node."""


class WebsocketNotConnectedError(NodeError):
    """Raised when the node websocket is not connected."""


class TrackError(PyLavError):
    """Base exception for Track errors."""


class ManagedLavalinkNodeError(NodeError):
    """Base Exception for Managed Lavalink Node Exceptions"""


class HTTPError(PyLavError):
    """Base exception for HTTP request errors."""


class Unauthorized(HTTPError):
    """Raised when a REST request fails due to an incorrect password."""


class InvalidTrack(TrackError):
    """Raised when an invalid track was passed."""


class NodeUnhealthy(ManagedLavalinkNodeError):
    """Exception Raised when the node health checks fail"""


class InvalidArchitectureError(ManagedLavalinkNodeError):
    """Error thrown when the Managed Lavalink node is started on an invalid arch."""


class ManagedLavalinkAlreadyRunningError(ManagedLavalinkNodeError):
    """Exception thrown when a managed Lavalink node is already running"""


class PortAlreadyInUseError(ManagedLavalinkNodeError):
    """Exception thrown when the port is already in use"""


class ManagedLinkStartAbortedUseExternal(ManagedLavalinkNodeError):
    """Exception thrown when the managed lavalink node is started but aborted"""


class ManagedLavalinkStartFailure(ManagedLavalinkNodeError):
    """Exception thrown when a managed Lavalink node fails to start"""


class ManagedLavalinkPreviouslyShutdownError(ManagedLavalinkNodeError):
    """Exception thrown when a managed Lavalink node already has been shutdown"""


class EarlyExitError(ManagedLavalinkNodeError):
    """some placeholder text I cannot be bothered to add a meaning message atm"""


class UnsupportedJavaError(ManagedLavalinkNodeError):
    """Exception thrown when a managed Lavalink node doesn't have a supported Java"""


class UnexpectedJavaResponseError(ManagedLavalinkNodeError):
    """Exception thrown when Java returns an unexpected response"""


class NoProcessFound(ManagedLavalinkNodeError):
    """Exception thrown when the managed node process is not found"""


class IncorrectProcessFound(ManagedLavalinkNodeError):
    """Exception thrown when the managed node process is incorrect"""


class TooManyProcessFound(ManagedLavalinkNodeError):
    """Exception thrown when zombie processes are suspected"""


class LavalinkDownloadFailed(ManagedLavalinkNodeError, RuntimeError):
    """Downloading the Lavalink jar failed.

    Attributes
    ----------
    response : aiohttp.ClientResponse
        The response from the server to the failed GET request.
    should_retry : bool
        Whether the lib should retry downloading the jar.
    """

    def __init__(self, *args, response: aiohttp.ClientResponse, should_retry: bool = False):
        super().__init__(*args)
        self.response = response
        self.should_retry = should_retry

    def __repr__(self) -> str:
        str_args = [*map(str, self.args), self._response_repr()]
        return f"LavalinkDownloadFailed({', '.join(str_args)}"

    def __str__(self) -> str:
        return f"{super().__str__()} {self._response_repr()}"

    def _response_repr(self) -> str:
        return f"[{self.response.status} {self.response.reason}]"


class TrackNotFound(TrackError):
    """Raised when a track is not found."""


class CogAlreadyRegistered(PyLavError):
    """Raised when a cog is already registered."""


class CogHasBeenRegistered(PyLavError):
    """Raised when a cog is registered."""


class NoNodeAvailable(NodeError):
    """Raised when no node is available."""


class NoNodeWithRequestFunctionalityAvailable(NodeError):
    """Raised when no node with request functionality is available."""

    def __init__(self, message: str, feature: str, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.message = message
        self.feature = feature


class PlaylistError(PyLavError):
    """Base class for playlist related errors."""


class InvalidPlaylist(PlaylistError):
    """Raised when a playlist is invalid."""



from __future__ import annotations

import asyncio
from typing import TYPE_CHECKING, Any
from uuid import uuid4

import aiohttp
import asyncstdlib
import ujson

from pylav._logging import getLogger
from pylav.constants import REGION_TO_COUNTRY_COORDINATE_MAPPING, SUPPORTED_SOURCES
from pylav.events import Event
from pylav.exceptions import Unauthorized
from pylav.filters import ChannelMix, Distortion, Equalizer, Karaoke, LowPass, Rotation, Timescale, Vibrato, Volume
from pylav.filters.tremolo import Tremolo
from pylav.location import distance
from pylav.sql.models import NodeModel
from pylav.types import LavalinkResponseT, TrackT
from pylav.utils import AsyncIter

if TYPE_CHECKING:
    from pylav.node_manager import NodeManager
    from pylav.player import Player
    from pylav.query import Query
    from pylav.websocket import WebSocket


LOGGER = getLogger("PyLav.Node")


class Penalty:
    """Represents the penalty of the stats of a Node."""

    __slots__ = ("_stats",)

    def __init__(self, stats: Stats):
        self._stats = stats

    @property
    def player_penalty(self) -> int:
        """The penalty of the players playing on the node.

        This is the number of players playing in the node.
        """
        return self._stats.playing_players

    @property
    def cpu_penalty(self) -> float:
        """The penalty of the cpu load of the node."""
        return 1.05 ** (100 * self._stats.system_load) * 10 - 10

    @property
    def null_frame_penalty(self) -> float | int:
        """The penalty of the nulled frames of the node."""
        null_frame_penalty = 0
        if self._stats.frames_nulled != -1:
            null_frame_penalty = (1.03 ** (500 * (self._stats.frames_nulled / 3000))) * 300 - 300
            null_frame_penalty *= 2
        return null_frame_penalty

    @property
    def deficit_frame_penalty(self) -> float | int:
        """The penalty of the deficit frames of the node."""
        return (
            1.03 ** (500 * (self._stats.frames_deficit / 3000)) * 600 - 600 if self._stats.frames_deficit != -1 else 0
        )

    @property
    def total(self) -> float:
        """The total penalty of the node.

        This is the sum of the penalties of the node.
        """
        return self.player_penalty + self.cpu_penalty + self.null_frame_penalty + self.deficit_frame_penalty


class Stats:
    """Represents the stats of Lavalink node."""

    __slots__ = (
        "_node",
        "_data",
        "_penalty",
        "_memory",
        "_cpu",
        "_frame_stats",
    )

    def __init__(self, node: Node, data: dict):
        self._node = node
        self._data = data
        self._memory = data["memory"]
        self._cpu = data["cpu"]
        self._frame_stats = data.get("frameStats", {})
        self._penalty = Penalty(self)

    @property
    def uptime(self) -> int:
        """How long the node has been running for in milliseconds."""
        return self._data["uptime"]

    @property
    def uptime_seconds(self) -> float:
        """How long the node has been running for in seconds."""
        return self.uptime / 1000

    @property
    def players(self) -> int:
        """The amount of players connected to the node."""
        return self._data["players"]

    @property
    def playing_players(self) -> int:
        """The amount of players that are playing in the node."""
        return self._data["playingPlayers"]

    @property
    def memory_free(self) -> int:
        """The amount of memory free to the node."""
        return self._memory["free"]

    @property
    def memory_used(self) -> int:
        """The amount of memory that is used by the node."""
        return self._memory["used"]

    @property
    def memory_allocated(self) -> int:
        """The amount of memory allocated to the node."""
        return self._memory["allocated"]

    @property
    def memory_reservable(self) -> int:
        """The amount of memory reservable to the node."""
        return self._memory["reservable"]

    @property
    def cpu_cores(self) -> int:
        """The amount of cpu cores the system of the node has."""
        return self._cpu["cores"]

    @property
    def system_load(self) -> float:
        """The overall CPU load of the system."""
        return self._cpu["systemLoad"]

    @property
    def lavalink_load(self) -> float:
        """The CPU load generated by Lavalink."""
        return self._cpu["lavalinkLoad"]

    @property
    def frames_sent(self) -> int:
        """The number of frames sent to Discord.
        Warning
        -------
        Given that audio packets are sent via UDP, this number may not be 100% accurate due to dropped packets.
        """
        return self._frame_stats.get("sent", -1)

    @property
    def frames_nulled(self) -> int:
        """The number of frames that yielded null, rather than actual data."""
        return self._frame_stats.get("nulled", -1)

    @property
    def frames_deficit(self) -> int:
        """The number of missing frames. Lavalink generates this figure by calculating how many packets to expect
        per minute, and deducting ``frames_sent``. Deficit frames could mean the CPU is overloaded, and isn't
        generating frames as quickly as it should be.
        """
        return self._frame_stats.get("deficit", -1)

    @property
    def penalty(self) -> Penalty:
        """The penalty for the node."""
        return self._penalty


class Node:
    """Represents a Node connection with Lavalink.

    Note
    ----
    Nodes are **NOT** meant to be added manually, but rather with :func:`Client.add_node`.

    Attributes
    ----------
    host: :class:`str`
        The address of the Lavalink node.
    port: Optional[:class:`int`]
        The port to use for websocket and REST connections.
    password: :class:`str`
        The password used for authentication.
    name: :class:`str`
        The name the :class:`Node` is identified by.
    ssl: :class:`bool`
        Whether to use a ssl connection.
    """

    _config: NodeModel

    def __init__(
        self,
        manager: NodeManager,
        host: str,
        password: str,
        resume_key: str,
        resume_timeout: int,
        port: int | None = None,
        name: str | None = None,
        reconnect_attempts: int = -1,
        ssl: bool = False,
        search_only: bool = False,
        unique_identifier: int = None,
        disabled_sources: list[str] = None,
        managed: bool = False,
        extras: dict = None,
    ):
        from pylav.query import Query

        self._query_cls: Query = Query  # type: ignore
        self._manager = manager
        self._session = aiohttp.ClientSession(timeout=aiohttp.ClientTimeout(total=30), json_serialize=ujson.dumps)
        if unique_identifier is None:
            unique_identifier = str(uuid4())
        self._managed = managed
        self._region = None
        self._host = host
        self._name = name or f"{self._region}-{self._host}-{unique_identifier}"
        self._extras = extras or {}
        self._disabled_sources = set(disabled_sources or [])
        self._config = None  # type: ignore

        if self._manager.get_node_by_id(unique_identifier) is not None:
            raise ValueError(f"A Node with identifier:{unique_identifier} already exists.")
        self._identifier = unique_identifier
        self._ssl = ssl
        if port is None:
            self._port = 443 if self._ssl else 80
        else:
            self._port = port
        self._password = password

        self._resume_key = resume_key or self._identifier
        self._resume_timeout = resume_timeout
        self._reconnect_attempts = reconnect_attempts
        self._search_only = search_only
        self._sources = self._capabilities = set()
        self._coordinates = (0, 0)

        self._stats = None
        from pylav.websocket import WebSocket

        self._ready = asyncio.Event()
        self._ws = WebSocket(
            node=self,
            host=self.host,
            port=self.port,
            password=self.password,
            resume_key=self.resume_key,
            resume_timeout=self.resume_timeout,
            reconnect_attempts=self.reconnect_attempts,
            ssl=self.ssl,
        )

    @property
    def is_ready(self) -> bool:
        return self._ready.is_set() and self._ws.connected

    @property
    def coordinates(self) -> tuple[int, int]:
        """The coordinates of the node.

        Returns
        -------
        :class:`tuple`
            The coordinates of the node.
        """
        return self._coordinates

    @property
    def managed(self) -> bool:
        return self._managed

    @property
    def config(self) -> NodeModel:
        return self._config

    @property
    def identifier(self) -> int:
        """
        The identifier of the :class:`Node`.
        """
        return self._identifier

    @property
    def search_only(self) -> bool:
        return self._search_only

    @property
    def session(self) -> aiohttp.ClientSession:
        return self._session

    @property
    def websocket(self) -> WebSocket:
        """The websocket of the node."""
        return self._ws

    @property
    def node_manager(self) -> NodeManager:
        """The :class:`NodeManager` this node belongs to."""
        return self._manager

    @property
    def port(self) -> int:
        """The port of the node."""
        return self._port

    @property
    def ssl(self) -> bool:
        """Whether the node is using a ssl connection."""
        return self._ssl

    @property
    def connection_protocol(self) -> str:
        """The protocol used for the connection."""
        return "https" if self.ssl else "http"

    @property
    def host(self) -> str:
        """The host of the node."""
        return self._host

    @property
    def region(self) -> str:
        """The region of the node."""
        return self._region

    @property
    def name(self) -> str:
        """The name of the node."""
        return self._name

    @property
    def password(self) -> str:
        """The password of the node."""
        return self._password

    @property
    def resume_key(self) -> str:
        """The resume key of the node."""
        return self._resume_key

    @property
    def resume_timeout(self) -> int:
        """The timeout to use for resuming."""
        return self._resume_timeout

    @property
    def reconnect_attempts(self) -> int:
        """The number of attempts to reconnect to the node."""
        return self._reconnect_attempts

    @property
    def stats(self) -> Stats:
        """The stats of the node."""
        return self._stats

    @stats.setter
    def stats(self, value: Stats) -> None:
        if not isinstance(value, Stats):
            raise TypeError("stats must be of type Stats")
        self._stats = value

    @property
    def available(self) -> bool:
        """Returns whether the node is available for requests."""
        return self._ws.connected if self._ws else False

    @property
    def _original_players(self) -> list[Player]:
        """Returns a list of players that were assigned to this node, but were moved due to failover etc."""
        return [p for p in self._manager.client.player_manager.players.values() if p._original_node == self]

    @property
    def players(self) -> list[Player]:
        """Returns a list of all players on this node."""
        return [p for p in self._manager.client.player_manager.players.values() if p.node == self]

    @property
    def playing_players(self) -> list[Player]:
        """Returns a list of all players on this node that are playing."""
        return [p for p in self.players if p.is_playing]

    @property
    def connected_players(self) -> list[Player]:
        """Returns a list of all players on this node that are connected."""
        return [p for p in self.players if p.is_connected]

    @property
    def server_connected_players(self) -> int:
        """Returns the number of players on this node that are connected."""
        return self.stats.players if self.stats else len(self.connected_players)

    @property
    def server_playing_players(self) -> int:
        """Returns the number of players on this node that are playing."""
        return self.stats.playing_players if self.stats else len(self.playing_players)

    @property
    def count(self) -> int:
        """Returns the number of players on this node."""
        return len(self.players)

    @property
    def playing_count(self) -> int:
        """Returns the number of players on this node that are playing."""
        return len(self.playing_players)

    @property
    def connected_count(self) -> int:
        """Returns the number of players on this node that are connected."""
        return len(self.connected_players)

    @property
    def penalty(self) -> float:
        """Returns the load-balancing penalty for this node."""
        if not self.available or not self.stats:
            return float("inf")
        return self.stats.penalty.total

    async def penalty_with_region(self, region: str | None) -> float:
        """The penalty for the node, with the region added in."""
        if not region:
            return self.penalty
        return self.penalty + (1.1 ** (0.0025 * await self.region_distance(region)) * 500 - 500)

    def dispatch_event(self, event: Event) -> None:
        """|coro|
        Dispatches the given event to all registered hooks.
        Parameters
        ----------
        event: :class:`Event`
            The event to dispatch to the hooks.
        """
        self.node_manager.client.dispatch_event(event)

    async def send(self, **data: Any) -> None:
        """|coro|
        Sends the passed data to the node via the websocket connection.
        Parameters
        ----------
        data: class:`any`
            The dict to send to Lavalink.
        """
        await self.websocket.send(**data)

    def __repr__(self):
        return (
            f"<Node id={self.identifier} name={self.name} "
            f"region={self.region} ssl={self.ssl} search_only={self.search_only} status={self._ws.connected}>"
        )

    def __eq__(self, other):
        if isinstance(other, Node):
            return self.identifier == other.identifier
        elif isinstance(other, NodeModel):
            return self.host == other.yaml["server"]["address"] and self.port == other.yaml["server"]["port"]
        return NotImplemented

    def __ne__(self, other):
        if isinstance(other, Node):
            return self.identifier != other.identifier
        elif isinstance(other, NodeModel):
            return not (self.host == other.yaml["server"]["address"] and self.port == other.yaml["server"]["port"])
        return NotImplemented

    def __hash__(self) -> int:
        return hash((self.host, self.port))

    async def get_query_youtube_music(self, query: str, bypass_cache: bool = False) -> LavalinkResponseT:
        """|coro|
        Gets the query from YouTube music.
        Parameters
        ----------
        query: :class:`str`
            The query to search for.
        bypass_cache: :class:`bool`
            Whether to bypass the cache.
        Returns
        -------
        list[dict]
            The list of results.
        """
        if not self.available:
            return {
                "loadType": "LOAD_FAILED",
                "playlistInfo": {"name": "", "selectedTrack": -1},
                "tracks": [],
            }
        query = f"ytmsearch:{query}"
        return await self.get_tracks(await self._query_cls.from_string(query), bypass_cache=bypass_cache)

    async def get_query_youtube(self, query: str, bypass_cache: bool = False) -> LavalinkResponseT:
        """|coro|
        Gets the query from YouTube music.
        Parameters
        ----------
        query: :class:`str`
            The query to search for.
        bypass_cache: :class:`bool`
            Whether to bypass the cache.
        Returns
        -------
        list[dict]
            The list of results.
        """
        if not self.available:
            return {
                "loadType": "LOAD_FAILED",
                "playlistInfo": {"name": "", "selectedTrack": -1},
                "tracks": [],
            }
        query = f"ytsearch:{query}"
        return await self.get_tracks(await self._query_cls.from_string(query), bypass_cache=bypass_cache)

    async def get_query_soundcloud(self, query: str, bypass_cache: bool = False) -> LavalinkResponseT:
        """|coro|
        Gets the query from Soundcloud.
        Parameters
        ----------
        query: :class:`str`
            The query to search for.
        bypass_cache: :class:`bool`
            Whether to bypass the cache.
        Returns
        -------
        list[dict]
            The list of results.
        """
        if not self.available:
            return {
                "loadType": "LOAD_FAILED",
                "playlistInfo": {"name": "", "selectedTrack": -1},
                "tracks": [],
            }
        query = f"scsearch:{query}"
        return await self.get_tracks(await self._query_cls.from_string(query), bypass_cache=bypass_cache)

    async def get_query_speak(self, query: str, bypass_cache: bool = False) -> list | None:
        """|coro|
        Gets the query for speak.
        Parameters
        ----------
        query: :class:`str`
            The query to search for.
        bypass_cache: :class:`bool`
            Whether to bypass the cache.
        Returns
        -------
        list[dict]
            The list of results.
        """
        if not self.available:
            return []
        if len(query) > 200:
            query = query[:200]
        query = f"speak:{query}"
        response = await self.get_tracks(
            await self._query_cls.from_string(query), bypass_cache=bypass_cache, first=True
        )
        return response.get("tracks")

    async def get_query_spotify(self, query: str, bypass_cache: bool = False) -> LavalinkResponseT:
        """|coro|
        Gets the query from Spotify.
        Parameters
        ----------
        query: :class:`str`
            The query to search for.
        bypass_cache: :class:`bool`
            Whether to bypass the cache.

        Returns
        -------
        list[dict]
            The list of results.
        """
        if not self.available:
            return {
                "loadType": "LOAD_FAILED",
                "playlistInfo": {"name": "", "selectedTrack": -1},
                "tracks": [],
            }
        query = f"spsearch:{query}"
        return await self.get_tracks(await self._query_cls.from_string(query), bypass_cache=bypass_cache)

    async def get_query_apple_music(self, query: str, bypass_cache: bool = False) -> LavalinkResponseT:
        """|coro|
        Gets the query from Apple Music.
        Parameters
        ----------
        query: :class:`str`
            The query to search for.
        bypass_cache: :class:`bool`
            Whether to bypass the cache.
        Returns
        -------
        list[dict]
            The list of results.
        """
        if not self.available:
            return {
                "loadType": "LOAD_FAILED",
                "playlistInfo": {"name": "", "selectedTrack": -1},
                "tracks": [],
            }
        query = f"amsearch:{query}"
        return await self.get_tracks(await self._query_cls.from_string(query), bypass_cache=bypass_cache)

    async def get_query_localfiles(
        self, query: str, bypass_cache: bool = True, first: bool = True
    ) -> LavalinkResponseT | TrackT:
        """|coro|
        Gets the query from Localfiles.
        Parameters
        ----------
        query: :class:`str`
            The query to search for.
        bypass_cache: :class:`bool`
            Whether to bypass the cache.
        first: :class:`bool`
            Whether to return the first result only.

        Returns
        -------
        list[dict]
            The list of results.
        """
        return (
            await self.get_tracks(
                await self._query_cls.from_string(query),
                bypass_cache=bypass_cache,
                first=first,
            )
            if self.available
            else {
                "loadType": "LOAD_FAILED",
                "playlistInfo": {"name": "", "selectedTrack": -1},
                "tracks": [],
            }
        )

    async def get_tracks(
        self, query: Query, first: bool = False, bypass_cache: bool = False
    ) -> LavalinkResponseT | TrackT:
        """|coro|
        Gets all tracks associated with the given query.

        Parameters
        ----------
        query: :class:`Query`
            The query to perform a search for.
        first: :class:`bool`
            Whether to return the first result or all results.
        bypass_cache: :class:`bool`
            Whether to bypass the cache.
        Returns
        -------
        :class:`dict`
            A dict representing tracks.
        """
        if not bypass_cache and (response := await self.node_manager.client.query_cache_manager.get_query(query)):
            return (
                {"track": response.tracks[0]}
                if first
                else {
                    "playlistInfo": {
                        "name": response.name,
                    },
                    "loadType": "PLAYLIST_LOADED"
                    if query.is_playlist or query.is_album
                    else "SEARCH_RESULT"
                    if query.is_search
                    else "TRACK_LOADED",
                    "tracks": [{"track": track} async for track in AsyncIter(response.tracks)],
                }
            )

        destination = f"{self.connection_protocol}://{self.host}:{self.port}/loadtracks"
        async with self._session.get(
            destination, headers={"Authorization": self.password}, params={"identifier": query.query_identifier}
        ) as res:
            if res.status == 200:
                result = await res.json(loads=ujson.loads)
                asyncio.create_task(self.node_manager.client.query_cache_manager.add_query(query, result))
                if first:
                    return await asyncstdlib.anext(  # type:ignore
                        asyncstdlib.iter(result.get("tracks", [])), default={}
                    )
                return result
            if res.status in [401, 403]:
                raise Unauthorized
            return {}

    async def decode_track(self, track: str) -> TrackT | None:
        destination = f"{self.connection_protocol}://{self.host}:{self.port}/decodetrack"
        async with self.session.get(
            destination, headers={"Authorization": self.password}, params={"track": track}
        ) as res:
            if res.status == 200:
                return await res.json(loads=ujson.loads)

            if res.status in [401, 403]:
                raise Unauthorized

            return None

    async def decode_tracks(self, tracks: list[str]) -> list[TrackT]:
        destination = f"{self.connection_protocol}://{self.host}:{self.port}/decodetracks"
        async with self.session.get(destination, headers={"Authorization": self.password}, json=tracks) as res:
            if res.status == 200:
                return await res.json(loads=ujson.loads)

            if res.status in [401, 403]:
                raise Unauthorized
            return []

    async def routeplanner_status(self) -> dict | None:
        """|coro|
        Gets the route-planner status of the target node.

        Returns
        -------
        :class:`dict`
            A dict representing the route-planner information.
        """
        destination = f"{self.connection_protocol}://{self.host}:{self.port}/routeplanner/status"
        async with self._session.get(destination, headers={"Authorization": self.password}) as res:
            if res.status == 200:
                return await res.json(loads=ujson.loads)

            if res.status in [401, 403]:
                raise Unauthorized
            return None

    async def routeplanner_free_address(self, address: str) -> bool:
        """|coro|
        Gets the route-planner status of the target node.

        Parameters
        ----------
        address: :class:`str`
            The address to free.

        Returns
        -------
        :class:`bool`
            True if the address was freed, False otherwise.
        """
        destination = f"{self.connection_protocol}://{self.host}:{self.port}/routeplanner/free/address"

        async with self._session.post(
            destination, headers={"Authorization": self.password}, json={"address": address}
        ) as res:
            if res.status in [401, 403]:
                raise Unauthorized
            return res.status == 204

    async def routeplanner_free_all_failing(self) -> bool:
        """|coro|
        Gets the route-planner status of the target node.

        Returns
        -------
        :class:`bool`
            True if all failing addresses were freed, False otherwise.
        """
        destination = f"{self.connection_protocol}://{self.host}:{self.port}/routeplanner/free/all"

        async with self._session.post(destination, headers={"Authorization": self.password}) as res:
            if res.status in [401, 403]:
                raise Unauthorized
            return res.status == 204

    async def get_plugins(self) -> list[dict]:
        """|coro|
        Gets the plugins of the target node.

        Returns
        -------
        :class:`list` of :class:`dict`
            A dict representing the plugins.
        """
        destination = f"{self.connection_protocol}://{self.host}:{self.port}/plugins"
        async with self._session.get(destination, headers={"Authorization": self.password}) as res:
            if res.status == 200:
                return await res.json(loads=ujson.loads)

            if res.status in [401, 403]:
                raise Unauthorized
        return []

    async def get_sources(self) -> dict:
        """|coro|
        Gets the sources of the target node.

        Returns
        -------
        :class:`dict`
            A dict representing the sources.
        """
        destination = f"{self.connection_protocol}://{self.host}:{self.port}/sources"
        async with self._session.get(destination, headers={"Authorization": self.password}) as res:
            if res.status == 200:
                return await res.json(loads=ujson.loads)

            if res.status in [401, 403]:
                raise Unauthorized
        return {}

    async def filters(
        self,
        *,
        guild_id: int,
        volume: Volume = None,
        equalizer: Equalizer = None,
        karaoke: Karaoke = None,
        timescale: Timescale = None,
        tremolo: Tremolo = None,
        vibrato: Vibrato = None,
        rotation: Rotation = None,
        distortion: Distortion = None,
        low_pass: LowPass = None,
        channel_mix: ChannelMix = None,
    ):
        op = {
            "op": "filters",
            "guildId": str(guild_id),
        }
        if volume and volume.changed:
            op["volume"] = volume.get()
        if equalizer and equalizer.changed:
            op["equalizer"] = equalizer.get()
        if karaoke and karaoke.changed:
            op["karaoke"] = karaoke.get()
        if timescale and timescale.changed:
            op["timescale"] = timescale.get()
        if tremolo and tremolo.changed:
            op["tremolo"] = tremolo.get()
        if vibrato and vibrato.changed:
            op["vibrato"] = vibrato.get()
        if rotation and rotation.changed:
            op["rotation"] = rotation.get()
        if distortion and distortion.changed:
            op["distortion"] = distortion.get()
        if low_pass and low_pass.changed:
            op["lowPass"] = low_pass.get()
        if channel_mix and channel_mix.changed:
            op["channelMix"] = channel_mix.get()

        await self.send(**op)

    async def get_unsupported_features(self) -> set[str]:
        if not self._capabilities:
            await self.update_features()
        return SUPPORTED_SOURCES - self._capabilities

    async def update_features(self):
        """|coro|
        Updates the features of the target node.
        """
        # This is pending a pr being merged until it is it will not provide any useful info
        #    However once it is merged this should be the only method to get capabilities
        #    from the node.
        for source_origin, source_data in (await self.get_sources()).items():
            if source_origin == "defaults":
                for source_name, source_state in source_data.items():
                    if source_state:
                        self._sources.add(source_name)
            elif source_origin == "plugins":
                for _, plugin_data in source_data.items():
                    for source_name, source_state in plugin_data.items():
                        if source_state:
                            self._sources.add(source_name)
        if not self._sources:
            if self.managed:
                self._capabilities.add("local")
            # FIXME: Remove me when the PR upstream is merged
            # This only exists as the above does not provide any useful info currently.
            #    However once it is merged this should be removed as it is not a good way to assess capabilities.
            #    As even though a plugin may be enable the source it adds may be disabled.

            #   Since this assumes everything is enable is is bound to cause track exceptions to be thrown when
            #   a source required but assumed enabled is in actuality disabled.
            for feature in await self.get_plugins():
                if feature["name"] == "Topis-Source-Managers-Plugin":
                    self._capabilities.update(["spotify", "applemusic"])
                elif feature["name"] == "DuncteBot-plugin":
                    self._capabilities.update(
                        [
                            "getyarn",
                            "clypit",
                            "speak",
                            "pornhub",
                            "reddit",
                            "ocremix",
                            "tiktok",
                            "mixcloud",
                            "soundgasm",
                        ]
                    )
                elif feature["name"] == "Google Cloud TTS":
                    self._capabilities.update(
                        "gcloud-tts",
                    )
                elif feature["name"] == "sponsorblock":
                    self._capabilities.add(
                        "sponsorblock",
                    )
            self._capabilities.update(["youtube", "soundcloud", "twitch", "bandcamp", "vimeo", "http"])
        # Give that remove files will not play nice with local files lets disable it for all but the managed node
        #    While this locks out some remove nodes with the correct setup it ensures that we are consistently
        #    using the correct node which will support local files instead of trying and failing most of the time.
        if not self.managed:
            self._capabilities.discard("local")
        # If not setup says these should be disabled remove them to trick the node to think they are disabled
        if self._sources:
            self._sources.difference_update(self._disabled_sources)

    def has_source(self, source: str) -> bool:
        """
        Checks if the target node has the specified source.

        Parameters
        ----------
        source: :class:`str`
            The source to check.

        Returns
        -------
        :class:`bool`
            True if the target node has the specified source, False otherwise.
        """
        return source.lower() in self.sources

    has_capability = has_source

    async def update_disabled_sources(self, sources: set[str]) -> None:
        """
        Updates the disabled sources.

        Returns
        -------
        :class:`None`
        """
        if self.managed:
            return
        unsupported = await self.get_unsupported_features()
        if self._config is None:
            self._config = await self.node_manager.client.node_db_manager.get_node_config(self.identifier)
        currently_disabled = set(self.config.disabled_sources)
        unsupported = unsupported.union(currently_disabled).union(sources)
        self.config.disabled_sources = list(unsupported)
        self._disabled_sources = unsupported
        await self.config.save()

    @property
    def capabilities(self) -> set:
        """
        Returns the capabilities of the target node.

        Returns
        -------
        :class:`set`
            The capabilities of the target node.
        """
        return self._capabilities.copy()

    @property
    def disabled_sources(self) -> set:
        """
        Returns the disabled sources of the target node.

        Returns
        -------
        :class:`set`
            The disabled sources of the target node.
        """
        return self._disabled_sources.copy()

    @property
    def sources(self) -> set:
        """
        Returns the sources of the target node.

        Returns
        -------
        :class:`set`
            The sources of the target node.
        """
        return self._sources.copy()

    @property
    def supports_spotify(self) -> bool:
        """
        Checks if the target node supports Spotify.

        Returns
        -------
        :class:`bool`
            True if the target node supports Spotify, False otherwise.
        """
        return self.has_source("spotify")

    @property
    def supports_apple_music(self) -> bool:
        """
        Checks if the target node supports Apple Music.

        Returns
        -------
        :class:`bool`
            True if the target node supports Apple Music, False otherwise.
        """
        return self.has_source("applemusic")

    @property
    def supports_getyarn(self) -> bool:
        """
        Checks if the target node supports GetYarn.

        Returns
        -------
        :class:`bool`
            True if the target node supports GetYarn, False otherwise.
        """
        return self.has_source("getyarn")

    @property
    def supports_clypit(self) -> bool:
        """
        Checks if the target node supports ClypIt.

        Returns
        -------
        :class:`bool`
            True if the target node supports ClypIt, False otherwise.
        """
        return self.has_source("clypit")

    @property
    def supports_speak(self) -> bool:
        """
        Checks if the target node supports speak source.

        Returns
        -------
        :class:`bool`
            True if the target node supports speak, False otherwise.
        """
        return self.has_source("speak")

    @property
    def supports_tts(self) -> bool:
        """
        Checks if the target node supports Google Cloud TTS.

        Returns
        -------
        :class:`bool`
            True if the target node supports Google Cloud TTS, False otherwise.
        """
        return self.has_capability("gcloud-tts")

    @property
    def supports_pornhub(self) -> bool:
        """
        Checks if the target node supports PornHub.

        Returns
        -------
        :class:`bool`
            True if the target node supports PornHub, False otherwise.
        """
        return self.has_source("pornhub")

    @property
    def supports_reddit(self) -> bool:
        """
        Checks if the target node supports Reddit.

        Returns
        -------
        :class:`bool`
            True if the target node supports Reddit, False otherwise.
        """
        return self.has_source("reddit")

    @property
    def supports_ocremix(self) -> bool:
        """
        Checks if the target node supports OCRemix.

        Returns
        -------
        :class:`bool`
            True if the target node supports OCRemix, False otherwise.
        """
        return self.has_source("ocremix")

    @property
    def supports_mixcloud(self) -> bool:
        """
        Checks if the target node supports Mixcloud.

        Returns
        -------
        :class:`bool`
            True if the target node supports Mixcloud, False otherwise.
        """
        return self.has_source("mixcloud")

    @property
    def supports_tiktok(self) -> bool:
        """
        Checks if the target node supports TikTok.

        Returns
        -------
        :class:`bool`
            True if the target node supports TikTok, False otherwise.
        """
        return self.has_source("tiktok")

    @property
    def supports_youtube(self) -> bool:
        """
        Checks if the target node supports YouTube.

        Returns
        -------
        :class:`bool`
            True if the target node supports YouTube, False otherwise.
        """
        return self.has_source("youtube")

    @property
    def supports_bandcamp(self) -> bool:
        """
        Checks if the target node supports Bandcamp.

        Returns
        -------
        :class:`bool`
            True if the target node supports Bandcamp, False otherwise.
        """
        return self.has_source("bandcamp")

    @property
    def supports_soundcloud(self) -> bool:
        """
        Checks if the target node supports SoundCloud.

        Returns
        -------
        :class:`bool`
            True if the target node supports SoundCloud, False otherwise.
        """
        return self.has_source("soundcloud")

    @property
    def supports_twitch(self) -> bool:
        """
        Checks if the target node supports Twitch.

        Returns
        -------
        :class:`bool`
            True if the target node supports Twitch, False otherwise.
        """
        return self.has_source("twitch")

    @property
    def supports_vimeo(self) -> bool:
        """
        Checks if the target node supports Vimeo.

        Returns
        -------
        :class:`bool`
            True if the target node supports Vimeo, False otherwise.
        """
        return self.has_source("vimeo")

    @property
    def supports_http(self) -> bool:
        """
        Checks if the target node supports HTTP.

        Returns
        -------
        :class:`bool`
            True if the target node supports HTTP, False otherwise.
        """
        return self.has_source("http")

    @property
    def supports_local(self) -> bool:
        """
        Checks if the target node supports local files.

        Returns
        -------
        :class:`bool`
            True if the target node supports local files, False otherwise.
        """
        return self.has_source("local") and self.has_source("local")

    @property
    def supports_sponsorblock(self) -> bool:
        """
        Checks if the target node supports SponsorBlock.

        Returns
        -------
        :class:`bool`
            True if the target node supports SponsorBlock, False otherwise.
        """
        return self.has_capability("sponsorblock")

    async def close(self) -> None:
        """
        Closes the target node.
        """
        if self._ws is not None:
            await self._ws.close()
        await self.session.close()

    async def wait_until_ready(self, timeout: float | None = None):
        await asyncio.wait_for(self._ready.wait(), timeout=timeout)

    async def region_distance(self, region: str) -> float:
        """
        Returns the numeric representation of the distance between the target node and the given region.

        Parameters
        ----------
        region : :class:`str`
            The region to get the distance to.

        Returns
        -------
        :class:`float`
            The numeric representation of the distance between the target node and the given region.
        """
        coordinates = REGION_TO_COUNTRY_COORDINATE_MAPPING.get(region)

        if not (coordinates and self.coordinates):
            return float("inf")

        return distance(*self.coordinates, *coordinates)



import datetime
import os
import re

from sphinx.builders.gettext import GettextRenderer, I18nBuilder, MessageCatalogBuilder, ltz, should_write, timestamp
from sphinx.builders.html import StandaloneHTMLBuilder
from sphinx.environment.adapters.indexentries import IndexEntries
from sphinx.locale import __
from sphinx.util import status_iterator
from sphinx.util.osutil import ensuredir
from sphinx.writers.html5 import HTML5Translator


class DPYHTML5Translator(HTML5Translator):
    def visit_section(self, node):
        self.section_level += 1
        self.body.append(self.starttag(node, "section"))

    def depart_section(self, node):
        self.section_level -= 1
        self.body.append("</section>\n")

    def visit_table(self, node):
        self.body.append('<div class="table-wrapper">')
        super().visit_table(node)

    def depart_table(self, node):
        super().depart_table(node)
        self.body.append("</div>")


class DPYStandaloneHTMLBuilder(StandaloneHTMLBuilder):
    # This is mostly copy pasted from Sphinx.
    def write_genindex(self) -> None:
        # the total count of lines for each index letter, used to distribute
        # the entries into two columns
        genindex = IndexEntries(self.env).create_index(self, group_entries=False)
        indexcounts = [sum(1 + len(subitems) for _, (_, subitems, _) in entries) for _k, entries in genindex]

        genindexcontext = {
            "genindexentries": genindex,
            "genindexcounts": indexcounts,
            "split_index": self.config.html_split_index,
        }

        if self.config.html_split_index:
            self.handle_page("genindex", genindexcontext, "genindex-split.html")
            self.handle_page("genindex-all", genindexcontext, "genindex.html")
            for (key, entries), count in zip(genindex, indexcounts):
                ctx = {"key": key, "entries": entries, "count": count, "genindexentries": genindex}
                self.handle_page(f"genindex-{key}", ctx, "genindex-single.html")
        else:
            self.handle_page("genindex", genindexcontext, "genindex.html")


class DPYMessageCatalogBuilder(MessageCatalogBuilder):
    _ADMONITION_REGEX = re.compile(r"\.\.\s*[a-zA-Z\_-]+::")

    def finish(self) -> None:
        # Bypass MessageCatalogBuilder.finish
        I18nBuilder.finish(self)

        # This is mostly copy pasted from Sphinx
        # However, this allows
        context = {
            "version": self.config.version,
            "copyright": self.config.copyright,
            "project": self.config.project,
            "last_translator": self.config.gettext_last_translator,
            "language_team": self.config.gettext_language_team,
            "ctime": datetime.datetime.fromtimestamp(timestamp, ltz).strftime("%Y-%m-%d %H:%M%z"),
            "display_location": self.config.gettext_location,
            "display_uuid": self.config.gettext_uuid,
        }

        REGEX = self._ADMONITION_REGEX
        for textdomain, catalog in status_iterator(
            self.catalogs.items(),
            __("writing message catalogs... "),
            "darkgreen",
            len(self.catalogs),
            self.app.verbosity,
            lambda textdomain__: textdomain__[0],
        ):
            # noop if config.gettext_compact is set
            ensuredir(os.path.join(self.outdir, os.path.dirname(textdomain)))

            # Due to a bug in Sphinx where messages contain admonitions, this code makes it
            # so they're suppressed from the output to prevent the output and CI from breaking
            # This is quite a bandaid fix but it seems to work ok
            # See https://github.com/sphinx-doc/sphinx/issues/10334
            context["messages"] = [msg for msg in catalog if REGEX.search(msg.text) is None]

            content = GettextRenderer(template_path="_templates/gettext", outdir=self.outdir).render(
                "message.pot_t", context
            )

            pofn = os.path.join(self.outdir, f"{textdomain}.pot")
            if should_write(pofn, content):
                with open(pofn, "w", encoding="utf-8") as pofile:
                    pofile.write(content)


def add_custom_jinja2(app):
    env = app.builder.templates.environment
    env.tests["prefixedwith"] = str.startswith
    env.tests["suffixedwith"] = str.endswith


def add_builders(app):
    """This is necessary because RTD injects their own for some reason."""
    app.set_translator("html", DPYHTML5Translator, override=True)
    app.add_builder(DPYStandaloneHTMLBuilder, override=True)
    app.add_builder(DPYMessageCatalogBuilder, override=True)

    try:
        original = app.registry.builders["readthedocs"]
    except KeyError:
        pass
    else:
        injected_mro = tuple(
            base if base is not StandaloneHTMLBuilder else DPYStandaloneHTMLBuilder for base in original.mro()[1:]
        )
        new_builder = type(original.__name__, injected_mro, {"name": "readthedocs"})
        app.set_translator("readthedocs", DPYHTML5Translator, override=True)
        app.add_builder(new_builder, override=True)


def setup(app):
    add_builders(app)
    app.connect("builder-inited", add_custom_jinja2)
    return {"parallel_read_safe": True}



from __future__ import annotations

import os
import pathlib
from collections.abc import AsyncIterator
from typing import Final

from discord.utils import maybe_coroutine

from pylav import Query
from pylav.vendored import aiopath

try:
    from redbot.core.i18n import Translator

    _ = Translator("PyLavPlayer", pathlib.Path(__file__))
except ImportError:
    _ = lambda x: x

__FULLY_SUPPORTED_MUSIC: Final[tuple[str, ...]] = (".mp3", ".flac", ".ogg")
__PARTIALLY_SUPPORTED_MUSIC_EXT: tuple[str, ...] = (
    ".m3u",
    ".m4a",
    ".aac",
    ".ra",
    ".wav",
    ".opus",
    ".wma",
    ".ts",
    ".au",
    # These do not work
    # ".mid",
    # ".mka",
    # ".amr",
    # ".aiff",
    # ".ac3",
    # ".voc",
    # ".dsf",
)
__PARTIALLY_SUPPORTED_VIDEO_EXT: tuple[str, ...] = (
    ".mp4",
    ".mov",
    ".flv",
    ".webm",
    ".mkv",
    ".wmv",
    ".3gp",
    ".m4v",
    ".mk3d",  # https://github.com/Devoxin/lavaplayer
    ".mka",  # https://github.com/Devoxin/lavaplayer
    ".mks",  # https://github.com/Devoxin/lavaplayer
    # These do not work
    # ".vob",
    # ".mts",
    # ".avi",
    # ".mpg",
    # ".mpeg",
    # ".swf",
)
__PARTIALLY_SUPPORTED_EXTENSION = __PARTIALLY_SUPPORTED_MUSIC_EXT + __PARTIALLY_SUPPORTED_VIDEO_EXT

_ALL_EXTENSIONS = __FULLY_SUPPORTED_MUSIC + __PARTIALLY_SUPPORTED_EXTENSION

_ROOT_FOLDER: aiopath.AsyncPath = None  # type: ignore


class LocalFile:
    _ROOT_FOLDER: aiopath.AsyncPath | None = _ROOT_FOLDER

    def __init__(self, path: str | pathlib.Path | aiopath.AsyncPath):
        if self._ROOT_FOLDER is None:
            raise RuntimeError(
                _("Root folder not initialized, call Client.update_localtracks_folder(folder: str | pathlib.Path)")
            )
        self._path: aiopath.AsyncPath = aiopath.AsyncPath(path)
        self._parent = self._path.parent
        self.__init = False

    def __str__(self) -> str:
        return str(self._path)

    async def initialize(self) -> None:
        if self.__init:
            return
        self._path = await maybe_coroutine(self._path.absolute)
        self._path.relative_to(self.root_folder)
        self.__init = True

    @classmethod
    async def add_root_folder(cls, path: str | pathlib.Path | aiopath.AsyncPath, *, create: bool = True):
        global _ROOT_FOLDER
        _ROOT_FOLDER = cls._ROOT_FOLDER = aiopath.AsyncPath(path)
        if create:
            await _ROOT_FOLDER.mkdir(parents=True, exist_ok=True)

    @property
    def path(self) -> aiopath.AsyncPath:
        return self._path

    @property
    def root_folder(self) -> aiopath.AsyncPath:
        return self._ROOT_FOLDER

    @property
    def parent(self) -> aiopath.AsyncPath:
        return self._parent

    @property
    def name(self) -> str:
        return self._path.name if pathlib.Path(self._path).is_dir() else self._path.stem

    @property
    def extension(self) -> str:
        return self._path.suffix

    @property
    def initialized(self) -> bool:
        return self.__init

    async def to_string_user(self, length: int = None, name_only: bool = False, ellipsis: bool = False) -> str:
        path = await maybe_coroutine(self.path.absolute)
        if name_only:
            string = path.name if await self.path.is_dir() else path.stem
        else:
            root = await maybe_coroutine(self.root_folder.absolute)
            string = str(path).replace(str(root), "")
        if not string:
            return path.name if await self.path.is_dir() else path.stem

        temp_path = aiopath.AsyncPath(string)
        if length is not None:
            parts = list(temp_path.parts)
            parts_reversed = list(parts[::-1])
            if await temp_path.is_file():
                parts_reversed[0] = temp_path.name
            count = 0
            usable_parts = []
            for part in parts_reversed:
                if usable_parts and (count + len(part) + 1) > length:
                    break
                count += len(part) + 1
                usable_parts.append(part)
            string_list = usable_parts[::-1]
            if name_only:
                string = os.path.join(*string_list[-2:])  # Folder and file only
            else:
                string = os.path.join(*string_list)
            if string.startswith("/") or string.startswith("\\"):
                string = string[1:]
            if ellipsis and len(string) + 3 > length:
                string = f"...{string[3:].strip()}"
        return string

    async def files_in_folder(self, show_folders: bool = False) -> AsyncIterator[Query]:
        parent = self.path if await self.path.is_dir() else self.path.parent
        async for path in self._get_entries_in_folder(parent, show_folders=show_folders):
            yield path

    async def _get_entries_in_folder(
        self, folder: aiopath.AsyncPath, recursive: bool = False, show_folders: bool = False, folder_only: bool = False
    ) -> AsyncIterator[Query]:
        async for path in folder.iterdir():
            if await path.is_dir():
                if recursive:
                    yield await Query.from_string(path)
                    async for p in self._get_entries_in_folder(
                        path, recursive=recursive, show_folders=show_folders, folder_only=folder_only
                    ):
                        yield p
                elif show_folders and path.is_relative_to(self._ROOT_FOLDER):
                    yield await Query.from_string(path)
            elif not folder_only and await path.is_file():
                if path.suffix.lower() in _ALL_EXTENSIONS and path.is_relative_to(self._ROOT_FOLDER):
                    yield await Query.from_string(path)

    async def files_in_tree(self, show_folders: bool = False) -> AsyncIterator[Query]:
        parent = self.path if await self.path.is_dir() else self.path.parent
        async for path in self._get_entries_in_folder(parent, recursive=True, show_folders=show_folders):
            yield path

    async def folders_in_tree(self) -> AsyncIterator[Query]:
        parent = self.path if await self.path.is_dir() else self.path.parent
        async for path in self._get_entries_in_folder(parent, recursive=True, show_folders=True, folder_only=True):
            yield path



from __future__ import annotations

import io
from collections.abc import AsyncIterable
from contextlib import asynccontextmanager
from inspect import iscoroutinefunction
from pathlib import Path
from typing import IO, TYPE_CHECKING, AsyncContextManager, BinaryIO, Final, TextIO, Union, cast

from aiofile import AIOFile, BinaryFileWrapper, LineReader, TextFileWrapper
from anyio import AsyncFile, open_file

from pylav.vendored.aiopath.aiopath_310.types import FileMode

if TYPE_CHECKING:  # keep mypy quiet
    from pylav.vendored.aiopath.aiopath_310.path import AsyncPath


BEGINNING: Final[int] = 0
CHUNK_SIZE: Final[int] = 4 * 1_024

SEP: Final[str] = "\n"
ENCODING: Final[str] = "utf-8"
ERRORS: Final[str] = "replace"


Paths = Union["AsyncPath", Path, str]
FileData = bytes | str


class FileLike(IO):
    is_binary: bool


class TextFile(FileLike, TextIO):
    pass


class BinaryFile(FileLike, BinaryIO):
    pass


class IterableAIOFile(FileLike, AIOFile):
    def __init__(self, *args, errors: str | None = ERRORS, newline: str | None = SEP, **kwargs):
        super().__init__(*args, **kwargs)
        self._errors: str | None = errors
        self._newline: str | None = newline

        self._offset: int = 0
        self._mode: str = self.__open_mode

    def __aiter__(self) -> AsyncIterable[str]:
        encoding, errors, line_sep = self._get_options()

        return read_lines(
            self.name,
            line_sep,
            encoding=encoding,
            errors=errors,
        )

    def _set_offset(self, offset: int, data: FileData):
        self._offset = offset + len(data)

    def _get_options(self, encoding: str | None = None, errors: str | None = None) -> tuple[str, str, str]:
        encoding = encoding or self.encoding or ENCODING
        errors = errors or self._errors or ERRORS
        line_sep: str = self._newline or SEP

        return encoding, errors, line_sep

    async def read_text(self, encoding: str | None = None, errors: str | None = None) -> str:
        encoding, errors, line_sep = self._get_options(encoding, errors)

        return await read_full_file(self.name, line_sep, encoding=encoding, errors=errors)

    async def read(self, size: int = -1, offset: int | None = None) -> FileData:
        if offset is None:
            offset = self._offset

        data: FileData = await super().read(size, offset)
        self._set_offset(offset, data)

        return data

    async def write(self, data: FileData, offset: int | None = None):
        if offset is None:
            offset = self._offset

        await super().write(data, offset)
        self._set_offset(offset, data)


Handle = TextFileWrapper | BinaryFileWrapper | IterableAIOFile | AsyncFile


async def read_lines(
    path: Paths,
    line_sep: str = SEP,
    chunk_size: int = CHUNK_SIZE,
    offset: int = BEGINNING,
    encoding: str = ENCODING,
    errors: str = ERRORS,
    **kwargs,
) -> AsyncIterable[str]:
    if hasattr(path, "resolve"):
        if iscoroutinefunction(path.resolve):
            path = str(await path.resolve())

        else:
            path = str(path.resolve())

    path = cast(str, path)

    async with AIOFile(path, "rb") as handle:
        reader = LineReader(handle, line_sep=line_sep, chunk_size=chunk_size, offset=offset)

        while line := await reader.readline():
            yield line.decode(encoding, errors=errors)


async def read_full_file(
    path: Paths,
    line_sep: str = SEP,
    chunk_size: int = CHUNK_SIZE,
    offset: int = BEGINNING,
    encoding: str = ENCODING,
    errors: str = ERRORS,
    **kwargs,
) -> str:
    lines_gen = read_lines(
        path, line_sep=line_sep, chunk_size=chunk_size, offset=offset, encoding=encoding, errors=errors
    )

    with io.StringIO() as string:
        async for line in lines_gen:
            string.write(line)

        return string.getvalue()


@asynccontextmanager
async def get_handle(
    name: str,
    mode: FileMode = "r",
    buffering: int = -1,
    encoding: str | None = ENCODING,
    errors: str | None = ERRORS,
    newline: str | None = SEP,
) -> AsyncContextManager[Handle]:
    file: AsyncFile

    if "b" in mode:
        file = await open_file(name, mode)

    else:
        file = await open_file(
            name,
            mode,
            encoding=encoding,
            errors=errors,
            newline=newline,
        )

    yield file
    await file.aclose()



from __future__ import annotations

from typing import TYPE_CHECKING

from piccolo.table import create_db_tables

from pylav._config import CONFIG_DIR
from pylav._logging import getLogger
from pylav.sql import tables
from pylav.sql.models import BotVersion, LibConfigModel
from pylav.vendored import aiopath

if TYPE_CHECKING:
    from pylav.client import Client

LOGGER = getLogger("PyLav.LibConfigManager")


class LibConfigManager:
    def __init__(self, client: Client):
        self._client = client
        self._config_folder = CONFIG_DIR

    async def initialize(self) -> None:
        await self.create_tables()

    @property
    def client(self) -> Client:
        return self._client

    @staticmethod
    async def create_tables() -> None:
        await create_db_tables(
            tables.PlaylistRow,
            tables.LibConfigRow,
            tables.PlayerStateRow,
            tables.NodeRow,
            tables.QueryRow,
            tables.BotVersionRow,
            tables.PlayerRow,
            tables.EqualizerRow,
            tables.AioHttpCacheRow,
            if_not_exists=True,
        )

    async def get_config(
        self,
        config_folder,
        localtrack_folder,
        java_path,
        enable_managed_node,
        auto_update_managed_nodes,
        use_bundled_pylav_external,
        use_bundled_lava_link_external,
    ) -> LibConfigModel:
        return await LibConfigModel.get_or_create(
            id=1,
            bot=self.client.bot.user.id,
            config_folder=f"{config_folder}",
            localtrack_folder=f"{localtrack_folder}",
            java_path=java_path,
            enable_managed_node=enable_managed_node,
            auto_update_managed_nodes=auto_update_managed_nodes,
            use_bundled_pylav_external=use_bundled_pylav_external,
            use_bundled_lava_link_external=use_bundled_lava_link_external,
        )

    async def set_lib_config(
        self,
        config_folder: aiopath.AsyncPath | str,
        java_path: str,
        localtrack_folder: aiopath.AsyncPath | str,
        enable_managed_node: bool,
        auto_update_managed_nodes: bool,
    ) -> LibConfigModel:
        config_folder: aiopath.AsyncPath = aiopath.AsyncPath(config_folder)
        localtrack_folder: aiopath.AsyncPath = aiopath.AsyncPath(localtrack_folder)
        if await config_folder.is_file():
            raise ValueError("The config folder must be a directory.")
        if await config_folder.is_dir() and not await config_folder.exists():
            await config_folder.mkdir(parents=True, exist_ok=True)

        self._config_folder = config_folder
        return await LibConfigModel(
            id=1,
            bot=self.client.bot.user.id,
            config_folder=str(config_folder),
            java_path=java_path,
            enable_managed_node=enable_managed_node,
            auto_update_managed_nodes=auto_update_managed_nodes,
            localtrack_folder=str(localtrack_folder),
        ).save()

    async def get_bot_db_version(self) -> BotVersion:
        from pylav._config import __VERSION__

        bv = BotVersion(bot=self._client.bot.user.id, version=__VERSION__)
        await bv.get_or_create()
        return bv

    async def update_bot_dv_version(self, version: str) -> None:
        bv = BotVersion(bot=self._client.bot.user.id, version=version)
        await bv.save()



from __future__ import annotations

from pylav.filters.channel_mix import ChannelMix
from pylav.filters.distortion import Distortion
from pylav.filters.equalizer import Equalizer
from pylav.filters.karaoke import Karaoke
from pylav.filters.lowpass import LowPass
from pylav.filters.rotation import Rotation
from pylav.filters.timescale import Timescale
from pylav.filters.vibrato import Vibrato
from pylav.filters.volume import Volume

__all__ = ("ChannelMix", "Distortion", "Equalizer", "Karaoke", "LowPass", "Rotation", "Timescale", "Vibrato", "Volume")
